class RequestError(Exception):
    """Used generate error messages for the http responses"""

    status_code = 400

    def __init__(self, message, status_code=None, payload=None):
        """Constructor

        Args:
            message (str): the message to send to the client
            code (int): the http error code (optionnal)
            payload (dict): additionnal information (optionnal)
        """
        Exception.__init__(self)
        self.message = message
        if status_code is not None:
            self.status_code = status_code
        self.payload = payload

    def __str__(self):
        return "Error {}: {}".format(self.status_code, self.message)

    def to_dict(self):
        data = dict(self.payload or ())
        data["message"] = self.message
        return data


class InvalidValue(RequestError):
    """Used to indicate an incorrect value for a field"""

    status_code = 400

    def __init__(self, message):
        return super().__init__(message=message)


class PermissionError(RequestError):
    status_code = 403

    def __init__(self, message="Permission denied"):
        return super().__init__(message=message)


class UserTokenError(Exception):
    """Used to indicate a user could not be associated with a json web token"""

    pass


class ProcessingError(Exception):
    """Used to indicate the error in the processing of a criterion"""

    pass


class InferenceError(Exception):
    """Used to indicate an error in the inference of the MR-Sort model."""

    pass


class UploadError(Exception):
    """Used to indicate a problem on uploaded files"""

    pass


class GeometryTypeError(TypeError):
    """Used to indicate the a geometry does not have the correct type"""

    pass


class InvalidAuthenticationHeader(Exception):
    """Used to indicate that the authentication header is not valid"""

    pass


class ExternalRequestError(Exception):
    """Used to indicate an external request to third party server failed.

    :param url: request tried
    :param msg: error message
    :param request_type: type of request tried, defaults 'External'
    """

    request_type = "External"

    def __init__(self, url: str, msg: str = None, request_type: str = None):
        self.url = url
        self.msg = msg or ""
        if request_type is not None:
            self.request_type = request_type

    def __str__(self) -> str:
        return f"{self.request_type} request {self.url} failed: {self.msg}"


class CapabilitiesXMLParsingError(Exception):
    """Used to indicate a parsing error on a XML GetCapabilities response
    (WMS/WFS)
    """

    pass



"""This module gathers functions/classes to extract openAPI specification
from this API.
"""

import importlib.resources
import inspect
import logging
import re
import sys
from typing import Type

import yaml
from apispec import APISpec, BasePlugin
from apispec.ext.marshmallow import MarshmallowPlugin
from apispec.ext.marshmallow.common import (
    get_unique_schema_name,
    make_schema_key,
    resolve_schema_instance,
)
from flask import current_app as app
from flask_restful import Api, Resource
from marshmallow import Schema, fields

from ..resources import api
from ..resources.utils import type2openapi
from ..schemas import *  # noqa: F403 (import all schemas for processing)


def get_schemas_classes(
    namespace: str | None = None,
) -> dict[str, Type[BaseSchema]]:  # noqa: F405
    """Get all schema classes from :mod:`api`.

    :param namespace:
        namespace from which to extract specification
        Defaults: ``__name__``
    :return: schema classes as a dictionary
    """
    namespace = namespace or __name__
    res = {}
    for name, obj in inspect.getmembers(sys.modules[namespace]):
        if inspect.isclass(obj) and issubclass(obj, Schema) and obj != Schema:
            res[name.replace("Schema", "")] = obj
    return res


def enum_fields2properties(self, field, **kwargs):
    """Helper function to convert enum field to openAPI spec.

    :param field:
    :return: openAPI spec
    """
    res = {}
    if isinstance(field, fields.Enum):
        values = []
        for member in field.enum:
            values.append(member.name)
        res["type"] = "string"
        res["enum"] = values
    return res


def pluck_fields2properties(self, field, **kwargs) -> dict:
    """Helper function to convert pluck field to openAPI spec.

    It uses a reference on the plucked field instead of its type.

    :param field:
    :return: openAPI spec
    """
    res = {}
    if isinstance(field, fields.Pluck):
        nested = (
            field.nested
            if isinstance(field.nested, str)
            else field.nested.__name__
        )
        ref = nested.replace("Schema", "")
        ref_field = f"#/components/schemas/{ref}/properties/{field.only[0]}"
        if field.many:
            res["items"] = res.get("items", {})
            res["items"]["$ref"] = ref_field
        else:
            res["$ref"] = ref_field
    return res


def method_fields2properties(self, field, **kwargs) -> dict:
    """Helper function to convert method field to openAPI spec.

    It uses the `field.field` field as the actual field to convert.

    :param field:
    :return: openAPI spec
    """
    ret = {}
    if isinstance(field, FieldMethod):  # noqa: F405
        ret = self.field2property(field.field)
    return ret


class MyMarshmallowPlugin(MarshmallowPlugin):
    """This class is a modified version of the basic Marshmallow plugin.

    It is capable of handling :class:`api.schemas.OneOfSchema` schemas.
    """

    schema_names = {}

    def schema_helper(self, name, _, schema=None, **kwargs):
        """Definition helper that allows using a marshmallow
        :class:`Schema <marshmallow.Schema>` to provide OpenAPI
        metadata.

        It was modified to handle correctly :class:`api.schemas.OneOfSchema`
        schemas.

        :param type|Schema schema: A marshmallow Schema class or instance.
        """
        if schema is None:
            return None

        schema_instance = resolve_schema_instance(schema)

        schema_key = make_schema_key(schema_instance)
        self.warn_if_schema_already_in_spec(schema_key)
        assert self.converter is not None, "init_spec has not yet been called"
        self.converter.refs[schema_key] = name

        if isinstance(schema_instance, OneOfSchema):  # noqa: F405
            json_schema = {
                "oneOf": [],
                "discriminator": {
                    "propertyName": schema_instance.type_field,
                    "mapping": {},
                },
            }
            for subname, sch_init in schema_instance.type_schemas.items():
                sch = sch_init(openapi=True)
                subsch_key = make_schema_key(sch)
                if subsch_key not in self.converter.refs:
                    new_subname = self.schema_names[sch.__class__]
                    self.spec.components.schema(
                        get_unique_schema_name(
                            self.spec.components, new_subname
                        ),
                        schema=sch,
                    )
                sch_ref = (
                    f"#/components/schemas/{self.converter.refs[subsch_key]}"
                )
                json_schema["oneOf"].append({"$ref": sch_ref})
                json_schema["discriminator"]["mapping"][subname] = sch_ref
            return json_schema

        return self.converter.schema2jsonschema(schema_instance)


def deep_update(dico: dict, new_dico: dict):
    """Perform a deep update on the first dictionnary with content of second.

    It handles nested dictionary structures, and lists.
    Everything else is updated the regular way.

    :param dico: dictionnary to update
    :param new_dico: updated data as a dictionnary
    """
    for k, v in new_dico.items():
        if k not in dico:
            dico[k] = v
        elif (  # noqa: E721
            type(dico[k]) == list and type(v) == list
        ):  # Make sure to exclude strings
            dico[k] += v
        elif isinstance(dico[k], dict) and isinstance(v, dict):
            deep_update(dico[k], v)
        else:
            dico[k] = v


class FlaskRestfulPlugin(BasePlugin):
    """This class is a :mod:`apispec` plugin made to handle
    :class:`flask_restful.Resource`.
    """

    def init_spec(self, spec: APISpec) -> None:
        """Initialize plugin with APISpec object

        :param APISpec spec: APISpec object this plugin instance is attached to
        """
        self.spec = spec

    def path_helper(
        self,
        path: str | None = None,
        operations: dict | None = None,
        parameters: list[dict] | None = None,
        *,
        resource: Type[Resource] | None = None,
        api: Api | None,
    ):
        """Add path to openAPI specification.

        :param api: ref to API
        :param path: endpoint of path (unused here)
        :param operations: operations statically defined on path
        :param parameters: parameters statically defined on path
        :param resource: API resource
        :raises ValueError: if `resource` is not found in `api`
        :return: openAPI path specification
        """
        assert resource is not None
        assert api is not None

        parameters = [] or parameters
        operations = {} or operations

        for klass, endpoints, _ in api.resources:
            if klass == resource:
                res = self.resource2openapi(klass, endpoints[0])
                operations.update(res["operations"])
                parameters += res.get("parameters", [])
                return res["path"]
        raise ValueError(f"Resource {resource.__name__} not found in API")

    @staticmethod
    def extract_doc_strings(func):
        """Extract any openAPI useful information from operation doc-string.

        :param func: operation function (get, put, post, etc.)
        :return: operation openAPI specification
        """
        res = {}
        doc = func.__doc__
        if doc is None:
            return res
        res["description"] = ""
        lines = iter(doc.splitlines())
        # Read description
        while (line := next(lines, None)) is not None:
            line = line.lstrip()
            if line.startswith(":"):
                break
            else:
                res["description"] += "\n" + line
        res["description"] = res["description"].strip()
        # Read rest of doc string
        tree = []
        current_node = None
        while line is not None:
            line = line.strip()
            if line == "---":
                # Reaching yaml part => abort
                break
            if re.match(r"^:", line):
                header = re.findall(r"^:[^:]*:", line)[0]
                if header.startswith(":param"):
                    # Parsing a path parameter
                    description = line[len(header) :].strip()
                    type_, arg = (
                        re.sub(r"^:param ", "", header)
                        .rstrip()[:-1]
                        .split(" ")
                    )
                    current_node = ["param", arg, type_, description]
                elif header.startswith(":status"):
                    # Parsing a response status
                    description = line[len(header) :].strip()
                    status = re.findall(r"\d+", header)[0]
                    current_node = ["status", status, description]
                else:
                    current_node = None
                tree.append(current_node)
            elif re.match(r"^.. :quickref: ", line):
                header = re.findall(r"^.. :quickref: ", line)[0]
                quickref = line[len(header) :]
                tag, description = quickref.split(";")[:2]
                current_node = ["quickref", tag, description]
                tree.append(current_node)
            elif current_node is not None:
                current_node[-1] += " " + line
            line = next(lines, None)

        for node in tree:
            match node:
                case None:
                    continue
                case ["param", arg, type_, description]:
                    res["parameters"] = res.get("parameters", [])
                    res["parameters"].append(
                        {
                            "name": arg,
                            "in": "path",
                            "required": True,
                            "schema": {"type": type2openapi(type_)},
                            "description": description.rstrip().lstrip(),
                        }
                    )
                case ["status", status, description]:
                    res["responses"] = res.get("responses", {})
                    res["responses"][str(status)] = {
                        "description": description.lstrip().rstrip()
                    }
                case ["quickref", tag, description]:
                    res["tags"] = res.get("tags", [])
                    res["tags"].append(tag)
                    res["summary"] = description.rstrip().lstrip()
        return res

    def resource2openapi(self, resource: Type[Resource], endpoint: str):
        """Convert API resource to openAPI specication.

        It extract spec from doc-string and special decorators
        (those modifying `func.__apispec__`).

        :param resource:
        :param endpoint: path endpoint
        :return: _description_
        """
        res = {"path": endpoint, "operations": {}}
        for method in resource.methods:
            func = getattr(resource, method.lower())
            operation = FlaskRestfulPlugin.extract_doc_strings(func)
            deep_update(operation, self.apispec_attr2openapi(func) or {})
            operation["responses"] = operation.get("responses", {})
            if "200" not in operation["responses"]:
                operation["responses"]["200"] = {"description": ""}
            res["operations"][method.lower()] = operation

        groups = re.findall(r"<[^<]*>", endpoint)
        if len(groups) == 0:
            return res
        res["parameters"] = []
        for group in groups:
            type_, arg = re.sub(r"[<>]", "", group).split(":")
            res["parameters"].append(
                {
                    "name": arg,
                    "in": "path",
                    "required": True,
                    "schema": {"type": type2openapi(type_)},
                }
            )
            res["path"] = res["path"].replace(group, f"{{{arg}}}")
        return res

    @property
    def marshmallow_plugin(self) -> MarshmallowPlugin | None:
        """Return :class:`MarshmallowPlugin` plugin if it is used to build
        the specification.

        :return:
        """
        for plugin in self.spec.plugins:
            if isinstance(plugin, MarshmallowPlugin):
                return plugin
        return None

    def _find_schema(self, schema: Schema) -> str | None:
        """Find name under which a schema is registered.

        :param schema:
        :return: registered name or ``None``
        """
        key = make_schema_key(schema)
        return self.marshmallow_plugin.converter.refs.get(key, None)

    def apispec_attr2openapi(self, func) -> dict:
        """Extract operation openAPI spec from function decorators.

        It extract spec from special decorators
        (those modifying `func.__apispec__`).

        :param func: operation function (get, put, post, etc.)
        :return: operation openAPI spec
        """
        return {
            **self.apispec_parameters2openapi(func),
            **self.apispec_responses2openapi(func),
            **self.apispec_requestBody2openapi(func),
            **self.apispec_security2openapi(func),
        }

    def apispec_security2openapi(self, func) -> dict:
        """Extract operation security openAPI spec from function decorators.

        It extract spec from `func.__apispec__['security']`.

        :param func: operation function (get, put, post, etc.)
        :return: operation openAPI spec
        """
        if not hasattr(func, "__apispec__"):
            return {}
        apispec_attr: dict = func.__apispec__
        if "security" not in apispec_attr:
            return {}
        return {"security": apispec_attr.get("security", [])}

    def apispec_parameters2openapi(self, func) -> dict:
        """Extract operation parameters openAPI spec from function decorators.

        It extract spec from `func.__apispec__['parameters']`.

        :param func: operation function (get, put, post, etc.)
        :return: operation openAPI spec
        """
        if not hasattr(func, "__apispec__"):
            return {}
        apispec_attr: dict = func.__apispec__
        if "parameters" not in apispec_attr:
            return {}
        return {"parameters": apispec_attr.get("parameters", [])}

    def apispec_requestBody2openapi(self, func) -> dict:
        """Extract operation request body openAPI spec from function
        decorators.

        It extract spec from `func.__apispec__['requestBody']`.

        :param func: operation function (get, put, post, etc.)
        :return: operation openAPI spec
        """
        if not hasattr(func, "__apispec__"):
            return {}
        apispec_attr: dict = func.__apispec__
        if "requestBody" not in apispec_attr:
            return {}
        request = apispec_attr.get("requestBody", {})

        res = {}
        res["requestBody"] = {
            "description": request.get("description", ""),
            "content": request.get("content", {}),
            "required": request.get("required", True),
        }
        schema = request.get("schema", None)
        if isinstance(schema, str):
            deep_update(
                res["requestBody"]["content"],
                {"application/json": {"schema": {"$ref": schema}}},
            )
        elif isinstance(schema, list):
            names = []
            if self.marshmallow_plugin:
                for sch in schema:
                    schema_instance = resolve_schema_instance(sch)
                    name = self._find_schema(schema_instance)
                    if name is None:
                        name = get_unique_schema_name(
                            self.spec.components,
                            type(schema_instance).__name__,
                        )
                        self.spec.components.schema(
                            name, schema=schema_instance
                        )
                    names.append(name)
            deep_update(
                res["requestBody"]["content"],
                {
                    "application/json": {
                        "schema": {
                            "oneOf": [
                                {"$ref": f"#/components/schemas/{name}"}
                                for name in names
                            ]
                        }
                    }
                },
            )
        elif isinstance(schema, type) or isinstance(schema, Schema):
            schema_instance = resolve_schema_instance(schema)
            if self.marshmallow_plugin:
                name = self._find_schema(schema_instance)
                if name is None:
                    name = get_unique_schema_name(
                        self.spec.components, type(schema_instance).__name__
                    )
                    self.spec.components.schema(name, schema=schema_instance)
                if isinstance(schema, Schema) and schema.many:
                    deep_update(
                        res["requestBody"]["content"],
                        {
                            "application/json": {
                                "schema": {
                                    "type": "array",
                                    "items": {
                                        "$ref": f"#/components/schemas/{name}"
                                    },
                                }
                            }
                        },
                    )
                else:
                    deep_update(
                        res["requestBody"]["content"],
                        {
                            "application/json": {
                                "schema": {
                                    "$ref": f"#/components/schemas/{name}"
                                }
                            }
                        },
                    )
        return res

    def apispec_responses2openapi(self, func) -> dict:
        """Extract operation responses openAPI spec from function decorators.

        It extract spec from `func.__apispec__['responses']`.

        :param func: operation function (get, put, post, etc.)
        :return: operation openAPI spec
        """
        if not hasattr(func, "__apispec__"):
            return {}
        apispec_attr: dict = func.__apispec__
        if "responses" not in apispec_attr:
            return {}

        res = {}
        res["responses"] = {}
        for status, response in apispec_attr.get("responses", {}).items():
            res["responses"][str(status)] = {
                "description": response.get("description", ""),
                "content": response.get("content", {}),
            }
            schema = response.get("schema", None)
            if isinstance(schema, str):
                deep_update(
                    res["responses"][str(status)]["content"],
                    {"application/json": {"schema": {"$ref": schema}}},
                )
            elif isinstance(schema, list):
                names = []
                if self.marshmallow_plugin:
                    for sch in schema:
                        schema_instance = resolve_schema_instance(sch)
                        name = self._find_schema(schema_instance)
                        if name is None:
                            name = get_unique_schema_name(
                                self.spec.components,
                                type(schema_instance).__name__,
                            )
                            self.spec.components.schema(
                                name, schema=schema_instance
                            )
                        names.append(name)
                deep_update(
                    res["responses"][str(status)]["content"],
                    {
                        "application/json": {
                            "schema": {
                                "oneOf": [
                                    {"$ref": f"#/components/schemas/{name}"}
                                    for name in names
                                ]
                            }
                        }
                    },
                )
            elif isinstance(schema, type) or isinstance(schema, Schema):
                schema_instance = resolve_schema_instance(schema)
                if self.marshmallow_plugin:
                    name = self._find_schema(schema_instance)
                    if name is None:
                        name = get_unique_schema_name(
                            self.spec.components,
                            type(schema_instance).__name__,
                        )
                        self.spec.components.schema(
                            name, schema=schema_instance
                        )
                    if isinstance(schema, Schema) and schema.many:
                        deep_update(
                            res["responses"][str(status)]["content"],
                            {
                                "application/json": {
                                    "schema": {
                                        "type": "array",
                                        "items": {
                                            "$ref": (
                                                f"#/components/schemas/{name}"
                                            )
                                        },
                                    }
                                }
                            },
                        )
                    else:
                        deep_update(
                            res["responses"][str(status)]["content"],
                            {
                                "application/json": {
                                    "schema": {
                                        "$ref": f"#/components/schemas/{name}"
                                    }
                                }
                            },
                        )

        return res


def extract_openapi(namespace: str | None = None) -> APISpec:
    """Extract OpenAPI specification.

    :param namespace: namespace to limit to when importing schemas
    :return: specification
    """
    # Retrieve swagger config from app config
    swagger_kwargs = app.config.get("SWAGGER")
    API_ROOT = app.config.get("API_ROOT", "/api")
    SERVERS = [
        {
            "url": API_ROOT,
            "description": "Backend server",
        }
    ]
    TITLE = swagger_kwargs.get("title", "DESEASION API docs")
    VERSION = swagger_kwargs.get("version")
    OPENAPI_VERSION = swagger_kwargs.get("openapi", "3.0.2")

    logging.info("Extract openAPI specification")
    spec_path = importlib.resources.files(
        "deseasion.backend.openapi"
    ).joinpath("spec_geo_json.yml")
    with spec_path.open("r") as f:
        geo_json_spec = yaml.safe_load(f)

    ma_plugin = MyMarshmallowPlugin()
    plugins = [ma_plugin, FlaskRestfulPlugin()]
    spec = APISpec(
        TITLE,
        VERSION,
        OPENAPI_VERSION,
        servers=SERVERS,
        components=geo_json_spec.get("components", {}),
        plugins=plugins,
    )
    ma_plugin.converter.add_attribute_function(enum_fields2properties)
    ma_plugin.converter.add_attribute_function(method_fields2properties)

    schema_classes = get_schemas_classes(namespace)
    ma_plugin.schema_names = {klass: k for k, klass in schema_classes.items()}

    for k, klass in schema_classes.items():
        schema_instance = resolve_schema_instance(klass)
        schema_key = make_schema_key(schema_instance)
        if schema_key not in ma_plugin.converter.refs:
            spec.components.schema(
                get_unique_schema_name(spec.components, k), schema=klass
            )
    logging.debug(
        f"Schemas added:{len(spec.components.schemas)} found: "
        f"{len(schema_classes)}"
    )

    spec.components.security_scheme(
        "bearerAuth",
        {"type": "http", "scheme": "bearer", "bearerFormat": "JWT"},
    )
    spec.components.security_scheme(
        "basicAuth",
        {
            "type": "http",
            "scheme": "basic",
            "description": "Provide access token",
        },
    )
    spec.components.security_scheme(
        "refreshAuth",
        {
            "type": "http",
            "scheme": "bearer",
            "bearerFormat": "JWT",
            "description": "Provide refresh token",
        },
    )

    for klass, _, _ in api.resources:
        spec.path(resource=klass, api=api)

    return spec



"""Make data created from streams a project data only

Revision ID: 184b71fa9bdc
Revises: c6bfa5794b5c
Create Date: 2025-02-26 15:30:07.436792

"""

import sqlalchemy as sa
from alembic import op
from geoalchemy2 import Geometry
from sqlalchemy.dialects import postgresql

from deseasion.backend.services.utils import progress_bar

# revision identifiers, used by Alembic.
revision = "184b71fa9bdc"
down_revision = "c6bfa5794b5c"
branch_labels = None
depends_on = None


old_data_type = sa.Enum(
    "geo_data",
    "generator",
    "global_data",
    name="datatype",
)
new_data_type = sa.Enum(
    "geo_data",
    "generator",
    "global_data",
    "data_stream",
    name="datatype",
)
tmp_data_type = sa.Enum(
    "geo_data",
    "generator",
    "global_data",
    "data_stream",
    name="_datatype",
)

old_task_type = sa.Enum(
    "process_project_data",
    "process_zone_proposition",
    name="projecttasktype",
)
new_task_type = sa.Enum(
    "process_project_data",
    "process_zone_proposition",
    "update_stream",
    name="projecttasktype",
)
tmp_task_type = sa.Enum(
    "process_project_data",
    "process_zone_proposition",
    "update_stream",
    name="_projecttasktype",
)


base_data = sa.Table(
    "base_data",
    sa.MetaData(),
    sa.Column(
        "type",
        sa.Enum("geo_data", "generator", "global_data", name="datatype"),
        nullable=True,
    ),
    sa.Column("name", sa.String(), nullable=False),
    sa.Column("properties_modified_at", sa.DateTime(), nullable=True),
    sa.Column("created_at", sa.DateTime(), nullable=True),
    sa.Column("modified_at", sa.DateTime(), nullable=True),
    sa.Column("id", sa.Integer(), nullable=False),
    sa.Column(
        "extent",
        Geometry(
            spatial_index=False, from_text="ST_GeomFromEWKT", name="geometry"
        ),
        nullable=True,
    ),
)


old_geo_data = sa.Table(
    "geo_data",
    sa.MetaData(),
    sa.Column(
        "type",
        sa.Enum(
            "geo_data",
            "generated_geo_data",
            "proposition_geo_data",
            "global_data",
            name="geodatatype",
        ),
        nullable=True,
    ),
    sa.Column("id", sa.Integer(), nullable=False),
    sa.Column("name", sa.String(), nullable=False),
    sa.Column(
        "extent",
        Geometry(
            spatial_index=False,
            from_text="ST_GeomFromEWKT",
            name="geometry",
        ),
        nullable=True,
    ),
    sa.Column("properties_modified_at", sa.DateTime(), nullable=True),
    sa.Column("created_at", sa.DateTime(), nullable=True),
    sa.Column("modified_at", sa.DateTime(), nullable=True),
    sa.Column("original_name", sa.String(), nullable=True),
    sa.Column("source_driver", sa.String(), nullable=True),
    sa.Column("upload_user_id", sa.Integer(), nullable=True),
    sa.Column("description", sa.String(), nullable=True),
    sa.Column("stream_id", sa.Integer(), nullable=True),
    sa.Column("is_public", sa.Boolean(), nullable=False),
)

new_geo_data = sa.Table(
    "geo_data",
    sa.MetaData(),
    sa.Column(
        "type",
        sa.Enum(
            "geo_data",
            "generated_geo_data",
            "proposition_geo_data",
            "global_data",
            name="geodatatype",
        ),
        nullable=True,
    ),
    sa.Column("id", sa.Integer(), nullable=False),
    sa.Column("name", sa.String(), nullable=False),
    sa.Column(
        "extent",
        Geometry(
            spatial_index=False,
            from_text="ST_GeomFromEWKT",
            name="geometry",
        ),
        nullable=True,
    ),
    sa.Column("properties_modified_at", sa.DateTime(), nullable=True),
    sa.Column("created_at", sa.DateTime(), nullable=True),
    sa.Column("modified_at", sa.DateTime(), nullable=True),
    sa.Column("original_name", sa.String(), nullable=True),
    sa.Column("source_driver", sa.String(), nullable=True),
    sa.Column("upload_user_id", sa.Integer(), nullable=True),
    sa.Column("description", sa.String(), nullable=True),
    sa.Column("is_public", sa.Boolean(), nullable=False),
)

feature = sa.Table(
    "feature",
    sa.MetaData(),
    sa.Column("data_id", sa.Integer(), nullable=True),
    sa.Column(
        "explainability",
        postgresql.JSONB(astext_type=sa.Text()),
        nullable=True,
    ),
    sa.Column("id", sa.Integer(), nullable=False),
    sa.Column(
        "type",
        sa.Enum("feature", "geo_feature", name="featuretype"),
        nullable=True,
    ),
)

geo_feature = sa.Table(
    "geo_feature",
    sa.MetaData(),
    sa.Column("id", sa.Integer(), nullable=False),
    sa.Column(
        "geom",
        Geometry(
            spatial_index=False, from_text="ST_GeomFromEWKT", name="geometry"
        ),
        nullable=True,
    ),
)

attribute = sa.Table(
    "data_attribute",
    sa.MetaData(),
    sa.Column(
        "type",
        sa.Enum("quantitative", "nominal", "ordinal", name="attributetype"),
        nullable=True,
    ),
    sa.Column("name", sa.String(), nullable=False),
    sa.Column(
        "statistics",
        postgresql.JSONB(astext_type=sa.Text()),
        nullable=True,
    ),
    sa.Column("data_id", sa.Integer(), nullable=False),
    sa.Column("id", sa.Integer(), nullable=False),
    sa.Column("order", postgresql.JSONB(astext_type=sa.Text()), nullable=True),
)

data_value = sa.Table(
    "data_value",
    sa.MetaData(),
    sa.Column(
        "type",
        sa.Enum("quantitative", "nominal", "ordinal", name="attributetype"),
        nullable=True,
    ),
    sa.Column("attribute_id", sa.Integer(), nullable=True),
    sa.Column("feature_id", sa.Integer(), nullable=True),
    sa.Column("value", postgresql.JSONB(astext_type=sa.Text()), nullable=True),
    sa.Column("id", sa.Integer(), nullable=False),
)

old_project_data = sa.Table(
    "project_data",
    sa.MetaData(),
    sa.Column(
        "data_type",
        sa.Enum("geo_data", "generator", "global_data", name="datatype"),
        nullable=True,
    ),
    sa.Column("id", sa.Integer(), nullable=False),
    sa.Column("data_id", sa.Integer(), nullable=True),
    sa.Column("last_update", sa.DateTime(), nullable=True),
    sa.Column("project_id", sa.Integer(), nullable=False),
    sa.Column("description", sa.String(), nullable=True),
    sa.Column("name", sa.String(), nullable=True),
    sa.Column("active_model_id", sa.Integer(), nullable=True),
)

new_project_data = sa.Table(
    "project_data",
    sa.MetaData(),
    sa.Column(
        "data_type",
        sa.Enum("geo_data", "generator", "global_data", name="datatype"),
        nullable=True,
    ),
    sa.Column("id", sa.Integer(), nullable=False),
    sa.Column("data_id", sa.Integer(), nullable=True),
    sa.Column("last_update", sa.DateTime(), nullable=True),
    sa.Column("project_id", sa.Integer(), nullable=False),
    sa.Column("description", sa.String(), nullable=True),
    sa.Column("name", sa.String(), nullable=True),
    sa.Column("active_model_id", sa.Integer(), nullable=True),
    sa.Column("stream_id", sa.Integer(), nullable=True),
    sa.Column(
        "classes",
        postgresql.ARRAY(sa.Float(), dimensions=2),
        nullable=True,
    ),
    sa.Column("start", sa.Float(), nullable=True),
    sa.Column("step", sa.Float(), nullable=True),
    sa.Column("stop", sa.Float(), nullable=True),
    sa.Column("resolution", sa.Float(), nullable=True),
)

input_assoc = sa.Table(
    "data_input_association",
    sa.MetaData(),
    sa.Column(
        "project_data_id",
        sa.Integer,
        nullable=False,
    ),
    sa.Column(
        "input_data_id",
        sa.Integer,
        nullable=False,
    ),
)

feature_input_assoc = sa.Table(
    "feature_input_association",
    sa.MetaData(),
    sa.Column(
        "feature_id",
        sa.Integer,
        nullable=False,
    ),
    sa.Column(
        "input_feature_id",
        sa.Integer,
        nullable=False,
    ),
)

project = sa.Table(
    "project",
    sa.MetaData(),
    sa.Column("is_template", sa.Boolean(), nullable=False),
    sa.Column("name", sa.String(), nullable=False),
    sa.Column("manager_id", sa.Integer(), nullable=True),
    sa.Column("description", sa.String(), nullable=True),
    sa.Column(
        "extent",
        Geometry(
            spatial_index=False,
            from_text="ST_GeomFromEWKT",
            name="geometry",
            nullable=False,
        ),
        nullable=False,
    ),
    sa.Column("created_at", sa.DateTime(), nullable=True),
    sa.Column("modified_at", sa.DateTime(), nullable=True),
    sa.Column("id", sa.Integer(), nullable=False),
    sa.Column("is_public", sa.Boolean(), nullable=False),
)

stream_geo_data = sa.Table(
    "stream_geo_data",
    sa.MetaData(),
    sa.Column(
        "type", sa.Enum("wfs", "wms", name="streamdatatype"), nullable=True
    ),
    sa.Column("name", sa.String(), nullable=False),
    sa.Column("title", sa.String(), nullable=True),
    sa.Column("original_name", sa.String(), nullable=True),
    sa.Column("url", sa.String(), nullable=False),
    sa.Column("upload_user_id", sa.Integer(), nullable=True),
    sa.Column("description", sa.String(), nullable=True),
    sa.Column(
        "keywords",
        postgresql.ARRAY(sa.String(), dimensions=1),
        nullable=True,
    ),
    sa.Column("is_public", sa.Boolean(), nullable=False),
    sa.Column("created_at", sa.DateTime(), nullable=True),
    sa.Column("modified_at", sa.DateTime(), nullable=True),
    sa.Column("id", sa.Integer(), nullable=False),
)

wms_geo_data = sa.Table(
    "wms_geo_data",
    sa.MetaData(),
    sa.Column("id", sa.Integer(), nullable=False),
    sa.Column(
        "classes",
        postgresql.ARRAY(sa.Float(), dimensions=2),
        nullable=True,
    ),
    sa.Column("layer", sa.String(), nullable=False),
    sa.Column("resolution", sa.Float(), nullable=True),
    sa.Column("version", sa.String(), nullable=True),
    sa.Column("start", sa.Float(), nullable=True),
    sa.Column("step", sa.Float(), nullable=True),
    sa.Column("stop", sa.Float(), nullable=True),
)

stream_permission = sa.Table(
    "stream_geo_data_permission",
    sa.MetaData(),
    sa.Column("object_id", sa.Integer(), nullable=True),
    sa.Column("user_id", sa.Integer(), nullable=False),
    sa.Column("id", sa.Integer(), nullable=False),
)

geo_data_permission = sa.Table(
    "geo_data_permission",
    sa.MetaData(),
    sa.Column("object_id", sa.Integer(), nullable=True),
    sa.Column("user_id", sa.Integer(), nullable=False),
    sa.Column("id", sa.Integer(), nullable=False),
)

default_value = sa.Table(
    "default_value",
    sa.MetaData(),
    sa.Column("attribute_id", sa.Integer(), nullable=False),
    sa.Column("model_id", sa.Integer(), nullable=False),
    sa.Column("value", postgresql.JSONB(astext_type=sa.Text()), nullable=True),
)

processing_model = sa.Table(
    "processing_model",
    sa.MetaData(),
    sa.Column(
        "model_type",
        sa.Enum(
            "categories_rule",
            "continuous_rule",
            "geo_buffer",
            "mrsort",
            "weighted_sum",
            "merge_overlap",
            "dissolve_adjacent",
            name="modeltype",
        ),
        nullable=True,
    ),
    sa.Column("name", sa.String(), nullable=True),
    sa.Column("data_generator_id", sa.Integer(), nullable=False),
    sa.Column(
        "keep_overlap",
        sa.Enum("min", "max", "sum", "average", name="keepoverlap"),
        nullable=False,
    ),
    sa.Column("cut_to_extent", sa.Boolean(), nullable=True),
    sa.Column("id", sa.Integer(), nullable=False),
    sa.Column("radius", sa.BigInteger(), nullable=True),
    sa.Column("majority_threshold", sa.Float(), nullable=True),
    sa.Column("rule", sa.String(), nullable=True),
)

mr_sort_criterion = sa.Table(
    "mrsort_criterion",
    sa.MetaData(),
    sa.Column(
        "profiles",
        postgresql.ARRAY(sa.Float(), dimensions=1),
        nullable=True,
    ),
    sa.Column("weight", sa.Float(), nullable=True),
    sa.Column("maximize", sa.Boolean(), nullable=False),
    sa.Column("mrsort_id", sa.Integer(), nullable=False),
    sa.Column("attribute_id", sa.Integer(), nullable=False),
)

weighted_sum_operand = sa.Table(
    "weighted_sum_operand",
    sa.MetaData(),
    sa.Column("attribute_id", sa.Integer(), nullable=True),
    sa.Column("weight", sa.Float(), nullable=True),
    sa.Column("model_id", sa.Integer(), nullable=False),
    sa.Column("id", sa.Integer(), nullable=False),
)


def upgrade():
    # Replace old data type with new one
    # https://stackoverflow.com/questions/14845203/altering-an-enum-field-using-alembic  # noqa: E501
    tmp_data_type.create(op.get_bind(), checkfirst=False)
    op.execute(
        "ALTER TABLE project_data ALTER COLUMN data_type TYPE _datatype"
        " USING data_type::text::_datatype"
    )
    old_data_type.drop(op.get_bind(), checkfirst=False)
    new_data_type.create(op.get_bind(), checkfirst=False)
    op.execute(
        "ALTER TABLE project_data ALTER COLUMN data_type TYPE datatype"
        " USING data_type::text::datatype"
    )
    tmp_data_type.drop(op.get_bind(), checkfirst=False)

    # Replace old task type with new one
    tmp_task_type.create(op.get_bind(), checkfirst=False)
    op.execute(
        "ALTER TABLE project_task ALTER COLUMN type TYPE _projecttasktype"
        " USING type::text::_projecttasktype"
    )
    old_task_type.drop(op.get_bind(), checkfirst=False)
    new_task_type.create(op.get_bind(), checkfirst=False)
    op.execute(
        "ALTER TABLE project_task ALTER COLUMN type TYPE projecttasktype"
        " USING type::text::projecttasktype"
    )
    tmp_task_type.drop(op.get_bind(), checkfirst=False)

    with op.batch_alter_table("project_data", schema=None) as batch_op:
        batch_op.add_column(
            sa.Column("stream_id", sa.Integer(), nullable=True)
        )
        batch_op.add_column(
            sa.Column(
                "classes",
                postgresql.ARRAY(sa.Float(), dimensions=2),
                nullable=True,
            )
        )
        batch_op.add_column(sa.Column("start", sa.Float(), nullable=True))
        batch_op.add_column(sa.Column("step", sa.Float(), nullable=True))
        batch_op.add_column(sa.Column("stop", sa.Float(), nullable=True))
        batch_op.add_column(sa.Column("resolution", sa.Float(), nullable=True))
        batch_op.create_foreign_key(
            batch_op.f("fk_project_data_stream_id_stream_geo_data"),
            "stream_geo_data",
            ["stream_id"],
            ["id"],
        )

    conn = op.get_bind()

    nb_migration_ops = conn.execute(
        sa.text(
            "SELECT count(p.id) FROM project_data AS p "
            "JOIN geo_data AS g ON g.id = p.data_id "
            "WHERE g.stream_id IS NOT NULL"
        )
    ).fetchone()[0]
    op_it = 0
    progress_bar(op_it, nb_migration_ops)

    # Replace all DataGeo created from stream by DataStream
    res = conn.execute(
        sa.select(
            new_project_data.c["id"],
            old_geo_data.c["stream_id", "id"],
            project.c["id", "manager_id"],
            stream_geo_data.c["is_public", "type"],
        )
        .join(old_geo_data, old_geo_data.c.id == new_project_data.c.data_id)
        .join(project, project.c.id == new_project_data.c.project_id)
        .join(
            stream_geo_data, stream_geo_data.c.id == old_geo_data.c.stream_id
        )
        .where(old_geo_data.c.stream_id != None)  # noqa: E711
    )
    geo_data_modified = []
    new_stream_perms = []
    for (
        p_data_id,
        stream_id,
        data_id,
        proj_id,
        manager_id,
        is_public,
        stream_type,
    ) in res:
        if stream_type == "wms":
            classes, start, step, stop, resolution = conn.execute(
                sa.select(
                    wms_geo_data.c[
                        "classes", "start", "step", "stop", "resolution"
                    ]
                ).where(wms_geo_data.c.id == stream_id)
            ).one()
        else:
            classes = None
            start = None
            step = None
            stop = None
            resolution = None

        # Replace DataGeo with DataStream instance (in place to keep refs)
        conn.execute(
            new_project_data.update()
            .where(new_project_data.c.id == p_data_id)
            .values(
                data_type="data_stream",
                stream_id=stream_id,
                classes=classes,
                start=start,
                step=step,
                stop=stop,
                resolution=resolution,
            )
        )
        if data_id in geo_data_modified:
            # geo_data used elsewhere => duplicate it
            data_sel = sa.select(
                base_data.c[
                    "type",
                    "name",
                    "extent",
                    "properties_modified_at",
                    "created_at",
                    "modified_at",
                ],
            ).where(base_data.c.id == data_id)
            new_data_id = conn.execute(
                base_data.insert()
                .from_select(
                    [
                        "type",
                        "name",
                        "extent",
                        "properties_modified_at",
                        "created_at",
                        "modified_at",
                    ],
                    data_sel,
                )
                .returning(base_data.c.id)
            ).one()[0]

            # Replace data_id in DataStream project data
            conn.execute(
                new_project_data.update()
                .where(new_project_data.c.id == p_data_id)
                .values(data_id=new_data_id)
            )

            # Duplicate attributes
            old_attrs = conn.execute(
                sa.select(
                    attribute.c["id", "type", "name", "statistics", "order"]
                ).where(attribute.c.data_id == data_id)
            )
            attributes_map = {}
            for attr_id, type, name, stats, order in old_attrs:
                attributes_map[attr_id] = conn.execute(
                    attribute.insert()
                    .values(
                        type=type,
                        data_id=new_data_id,
                        name=name,
                        statistics=stats,
                        order=order,
                    )
                    .returning(attribute.c.id)
                ).one()[0]
                # Update attribute links
                # Update links in default_value table
                model_ids = [
                    _row[0]
                    for _row in conn.execute(
                        sa.select(
                            default_value.c["model_id"],
                        )
                        .join(
                            processing_model,
                            processing_model.c.id == default_value.c.model_id,
                        )
                        .join(
                            new_project_data,
                            new_project_data.c.id
                            == processing_model.c.data_generator_id,
                        )
                        .join(
                            project,
                            project.c.id == new_project_data.c.project_id,
                        )
                        .where(project.c.id == proj_id)
                        .where(default_value.c.attribute_id == attr_id)
                    )
                ]
                conn.execute(
                    default_value.update()
                    .where(default_value.c.model_id.in_(model_ids))
                    .where(default_value.c.attribute_id == attr_id)
                    .values(attribute_id=attributes_map[attr_id])
                )
                # Update links in mr_sort_criterion
                mr_sort_ids = [
                    _row[0]
                    for _row in conn.execute(
                        sa.select(mr_sort_criterion.c["mrsort_id"])
                        .join(
                            processing_model,
                            processing_model.c.id
                            == mr_sort_criterion.c.mrsort_id,
                        )
                        .join(
                            new_project_data,
                            new_project_data.c.id
                            == processing_model.c.data_generator_id,
                        )
                        .join(
                            project,
                            project.c.id == new_project_data.c.project_id,
                        )
                        .where(project.c.id == proj_id)
                        .where(mr_sort_criterion.c.attribute_id == attr_id)
                    )
                ]
                conn.execute(
                    mr_sort_criterion.update()
                    .where(mr_sort_criterion.c.mrsort_id.in_(mr_sort_ids))
                    .where(mr_sort_criterion.c.attribute_id == attr_id)
                    .values(attribute_id=attributes_map[attr_id])
                )
                # Update links in weighted_sum_operand table
                model_ids = [
                    _row[0]
                    for _row in conn.execute(
                        sa.select(weighted_sum_operand.c["id"])
                        .join(
                            processing_model,
                            processing_model.c.id
                            == weighted_sum_operand.c.model_id,
                        )
                        .join(
                            new_project_data,
                            new_project_data.c.id
                            == processing_model.c.data_generator_id,
                        )
                        .join(
                            project,
                            project.c.id == new_project_data.c.project_id,
                        )
                        .where(project.c.id == proj_id)
                        .where(weighted_sum_operand.c.attribute_id == attr_id)
                    )
                ]
                conn.execute(
                    weighted_sum_operand.update()
                    .where(weighted_sum_operand.c.id.in_(model_ids))
                    .values(attribute_id=attributes_map[attr_id])
                )

            # Duplicate features as well
            orig_feats = conn.execute(
                sa.select(feature.c["id"], geo_feature.c["geom"])
                .join(geo_feature, geo_feature.c.id == feature.c.id)
                .where(feature.c.data_id == data_id)
            )
            for f_id, geom in orig_feats:
                # Duplicate feature
                new_f_id = conn.execute(
                    feature.insert()
                    .values(data_id=new_data_id, type="geo_feature")
                    .returning(feature.c["id"])
                ).one()[0]
                # Duplicate geometry
                conn.execute(
                    geo_feature.insert().values(id=new_f_id, geom=geom)
                )
                # Duplicate values
                old_values = conn.execute(
                    sa.select(
                        data_value.c["type", "attribute_id", "value"]
                    ).where(data_value.c.feature_id == f_id)
                )
                for type, attr_id, value in old_values:
                    conn.execute(
                        data_value.insert().values(
                            type=type,
                            attribute_id=attributes_map[attr_id],
                            feature_id=new_f_id,
                            value=value,
                        )
                    )

            # We also need to replace data and its nested fields in every
            # ref of the project
            # attributes used in processing models
        else:
            # Replace geo data type with 'generated_geo_data'
            conn.execute(
                base_data.update()
                .where(base_data.c.id == data_id)
                .values(type="generated_geo_data")
            )
            # Remove geo_data permissions
            # (generated geo data don't have permissions)
            conn.execute(
                geo_data_permission.delete().where(
                    geo_data_permission.c.object_id == data_id
                )
            )
            # Remove geo_data entry (generated_geo_data are not in this table!)
            conn.execute(
                old_geo_data.delete().where(old_geo_data.c.id == data_id)
            )
            geo_data_modified.append(data_id)
        if not is_public:
            # Make sure project manager can access stream
            # Handle when project manager can access version but not stream
            new_stream_perms.append((stream_id, manager_id))
        op_it += 1
        progress_bar(op_it, nb_migration_ops)

    # Make sure project manager can access streams in their project
    _new_stream_perms = set(new_stream_perms)
    res = conn.execute(sa.select(stream_permission.c["object_id", "user_id"]))
    _new_stream_perms -= {(r[0], r[1]) for r in res.all()}
    new_stream_perms = sorted(
        set(_new_stream_perms), key=new_stream_perms.index
    )
    for stream_id, user_id in new_stream_perms:
        conn.execute(
            stream_permission.insert().values(
                object_id=stream_id, user_id=user_id
            )
        )

    with op.batch_alter_table("geo_data", schema=None) as batch_op:
        batch_op.drop_constraint(
            "fk_geo_data_stream_id_stream_geo_data", type_="foreignkey"
        )
        batch_op.drop_column("stream_id")


def downgrade():
    # Remove update_stream tasks
    op.execute(sa.text("DELETE FROM project_task WHERE type = update_stream"))

    # Replace new task type with old one
    tmp_task_type.create(op.get_bind(), checkfirst=False)
    op.execute(
        "ALTER TABLE project_task ALTER COLUMN type TYPE _projecttasktype"
        " USING type::text::_projecttasktype"
    )
    new_task_type.drop(op.get_bind(), checkfirst=False)
    old_task_type.create(op.get_bind(), checkfirst=False)
    op.execute(
        "ALTER TABLE project_task ALTER COLUMN type TYPE projecttasktype"
        " USING type::text::projecttasktype"
    )
    tmp_task_type.drop(op.get_bind(), checkfirst=False)

    with op.batch_alter_table("geo_data", schema=None) as batch_op:
        batch_op.add_column(
            sa.Column(
                "stream_id", sa.INTEGER(), autoincrement=False, nullable=True
            )
        )
        batch_op.create_foreign_key(
            "fk_geo_data_stream_id_stream_geo_data",
            "stream_geo_data",
            ["stream_id"],
            ["id"],
        )

    conn = op.get_bind()
    nb_migration_ops = conn.execute(
        sa.text(
            "SELECT count(id) FROM project_data "
            "WHERE data_type = 'data_stream'"
        )
    ).fetchone()[0]
    op_it = 0
    progress_bar(op_it, nb_migration_ops)

    # Replace all DataStream by DataGeo with GeoData ref
    res = conn.execute(
        sa.select(
            new_project_data.c["id", "stream_id", "data_id"],
            project.c["manager_id"],
            stream_geo_data.c["upload_user_id", "name", "type"],
        )
        .join(project, project.c.id == new_project_data.c.project_id)
        .join(
            stream_geo_data,
            stream_geo_data.c.id == new_project_data.c.stream_id,
        )
        .where(new_project_data.c.data_type == "data_stream")  # noqa: E711
    )
    for (
        p_data_id,
        stream_id,
        data_id,
        manager_id,
        upload_user_id,
        name,
        s_type,
    ) in res:
        # Replace DataStream instance with DataGeo (in place to keep refs)
        conn.execute(
            new_project_data.update()
            .where(new_project_data.c.id == p_data_id)
            .values(data_type="geo_data")
        )
        # Replace geo data type with 'geo_data' and add stream ref
        conn.execute(
            base_data.update()
            .where(base_data.c.id == data_id)
            .values(type="geo_data")
        )
        conn.execute(
            old_geo_data.insert().values(
                id=data_id,
                stream_id=stream_id,
                upload_user_id=upload_user_id,
                original_name=name,
                source_driver=s_type,
                is_public=False,
            )
        )
        # Add access to geo data to stream upload user and DataStream manager
        conn.execute(
            geo_data_permission.insert().values(
                object_id=data_id, user_id=manager_id
            )
        )
        conn.execute(
            geo_data_permission.insert().values(
                object_id=data_id, user_id=upload_user_id
            )
        )
        op_it += 1
        progress_bar(op_it, nb_migration_ops)

    with op.batch_alter_table("project_data", schema=None) as batch_op:
        batch_op.drop_constraint(
            batch_op.f("fk_project_data_stream_id_stream_geo_data"),
            type_="foreignkey",
        )
        batch_op.drop_column("resolution")
        batch_op.drop_column("stop")
        batch_op.drop_column("step")
        batch_op.drop_column("start")
        batch_op.drop_column("classes")
        batch_op.drop_column("stream_id")

    # Replace new data type with old one
    # https://stackoverflow.com/questions/14845203/altering-an-enum-field-using-alembic  # noqa: E501
    tmp_data_type.create(op.get_bind(), checkfirst=False)
    op.execute(
        "ALTER TABLE project_data ALTER COLUMN data_type TYPE _datatype"
        " USING data_type::text::_datatype"
    )
    new_data_type.drop(op.get_bind(), checkfirst=False)
    old_data_type.create(op.get_bind(), checkfirst=False)
    op.execute(
        "ALTER TABLE project_data ALTER COLUMN data_type TYPE datatype"
        " USING data_type::text::datatype"
    )
    tmp_data_type.drop(op.get_bind(), checkfirst=False)



from celery import chain, group
from flask import current_app as app
from flask import jsonify
from flask_restful import Resource, reqparse
from flask_restful.inputs import boolean

from ..exceptions import InferenceError, RequestError
from ..models import DataAttribute, DataType, MRSort, ProjectTaskModel, db
from ..models.project_data_utils import ProjectDataProcessingService
from ..schemas import (
    DataGeneratorSchema,
    DataGeoSchema,
    DataStreamSchema,
    MessageSchema,
    MRSortInferenceGetResponseSchema,
    MRSortInferenceSchema,
    ProcessingModelSchema,
    ProjectDataActiveModelPutResponseSchema,
    ProjectDataGetResponseSchema,
    ProjectDataListGetResponseSchema,
    ProjectDataModelChangePostRequestBodySchema,
    ProjectDataPostResponseSchema,
    ProjectDataReplaceInputsPostRequestBodySchema,
    ProjectDataSchema,
    ProjectGlobalDataSchema,
    ProjectSharedDataListGetResponseSchema,
    ProjectTaskSchema,
)
from ..services import processing_model_service, project_data_service
from ..services.auth_service import token_required
from .utils import (
    dump_data,
    get_json_content,
    with_query_arg,
    with_request_body,
    with_response,
)


class ProjectDataListAPI(Resource):
    def parse_ids(self, args):
        """Returns the list of integer ids parsed from the comma-separated
        query arguments"""
        ids_str = args.get("ids", None)
        try:
            ids = (
                [
                    int(data_id)
                    for data_id in ids_str.split(",")
                    if len(data_id) > 0
                ]
                if ids_str is not None
                else []
            )
        except ValueError:
            raise RequestError(
                "The ids must be integers separated by a comma", 400
            )
        if len(ids) == 0:
            raise RequestError("A list of the ids is required", 400)
        return ids

    def parse_fields(self, args):
        """Returns the list of string fields parsed from the comma-separated
        query arguments"""
        fields_str = args.get("fields", None)
        fields = fields_str.split(",") if fields_str is not None else None
        return fields

    @token_required
    @with_query_arg(
        "fields",
        schema={"type": "array", "items": {"type": "string"}},
        required=False,
        description="The list of project data fields to return. Default: all.",
        style="form",
        explode=False,
    )
    @with_query_arg(
        "ids",
        schema={"type": "array", "items": {"type": "integer"}},
        description="The list of project data ids",
        style="form",
        explode=False,
    )
    @with_response(
        200,
        ProjectDataListGetResponseSchema,
        description="List of project data with wanted fields",
    )
    def get(self):
        """
        Returns the list of project data

        :query ids: The list of comma separated data ids (**required**).
        :query fields:
            The list of fields to return. Default: Return all fields.
        :reqheader Authorization:
            JSON Web Token with Bearer scheme
            (``Authorization: Bearer <token>``).
        :status 403:
            The user is not allowed to access at least one project data.
        :status 404: At least one project data does not exist.

        .. :quickref: Project data; Get the list of project data
        """
        parser = reqparse.RequestParser()
        parser.add_argument("ids", type=str, location="args")
        parser.add_argument("fields", type=str, location="args")
        args = parser.parse_args()
        ids = self.parse_ids(args)
        fields = self.parse_fields(args)
        data_list = [
            project_data_service.get_if_authorized(data_id) for data_id in ids
        ]
        serialized = []
        for data in data_list:
            data_dict = ProjectDataSchema().dump(data)
            if fields is not None:
                data_dict = {k: v for k, v in data_dict.items() if k in fields}
            serialized.append(data_dict)
        return jsonify(project_data=serialized)


class ProjectDataAPI(Resource):
    def create_process_task(self, data_id):
        """Create the appropriate type of task to process data"""
        data = project_data_service.get_if_authorized(data_id)
        if data.data_type == DataType.data_stream:
            return processing_model_service.update_stream_process(data.id)
        else:
            return processing_model_service.task_process(
                data.processing_model.id
            )

    def get_chained_tasks(self, data):
        """Start the celery tasks"""
        data_id_groups = project_data_service.get_grouped_dependant_data_ids(
            data
        )
        tasks = []
        for data_id_group in data_id_groups:
            # TODO: Group process of data
            if len(data_id_group) == 1:
                tasks.append(self.create_process_task(data_id_group[0]))
            elif len(data_id_group) > 1:
                grouped = [
                    self.create_process_task(data_id)
                    for data_id in data_id_group
                ]
                tasks.append(group(*grouped))
        if len(tasks) == 1:
            task = tasks[0]
            app.logger.info("starting task: {}".format(task))
            return task
        elif len(tasks) > 1:
            chain_signature = chain(*tasks)
            app.logger.info(
                "starting chain of tasks: {}".format(chain_signature)
            )
            return chain_signature

    def process_data(self, data_id, chain=False):
        """Process the data in an asynchronous task

        Args:
            data_id (int): the id of the data to process
            chain (bool):
                also process the required dependencies (default: False)
        """
        # TODO: Repair chained processing
        data = project_data_service.get_if_authorized(data_id)
        if chain:
            task_s = self.get_chained_tasks(data)
        else:
            task_s = self.create_process_task(data.id)
        task = task_s.apply_async(countdown=1)
        # Temporary object which is not persisted
        # to send the task_id to the user
        temporary_task = ProjectTaskModel(
            task_id=task.id, project=data.project
        )
        return dump_data(ProjectTaskSchema(), task=temporary_task)

    @token_required
    @with_response(
        202,
        schema=ProjectDataPostResponseSchema,
        description="Details of the newly created asynchronously running task",
    )
    @with_query_arg(
        "chain",
        bool,
        required=False,
        description="Also process the required dependencies. Default: False.",
    )
    @with_query_arg(
        "action",
        str,
        description=(
            "Action to execute on the data: currently only 'process' is "
            "supported."
        ),
    )
    def post(self, data_id):
        """
        Process the project's data.

        :param int data_id: The id of the project's data.
        :query str action:
            Action to execute on the data: currently only ``process`` is
            supported (**required**).
        :query bool chain:
            Also process the required dependencies. Default: False.
        :reqheader Authorization:
            JSON Web Token with Bearer scheme
            (``Authorization: Bearer <token>``).
        :status 202:
            The data is being processed asynchronously.
            The detail of the running task is returned.
        :status 400: The ``action`` is incorrect.
        :status 403: The user is not allowed to access the project.
        :status 404: The data does not exist.

        .. :quickref: Project data; Process the project's data
        """
        parser = reqparse.RequestParser()
        parser.add_argument("action", type=str, location="args", required=True)
        parser.add_argument(
            "chain", type=boolean, location="args", default=False
        )
        args = parser.parse_args()
        if args.get("action") == "process":
            chain = args.get("chain")
            resp = self.process_data(data_id, chain)
            resp.status_code = 202
            return resp
        else:
            raise RequestError("Unknown action")

    @token_required
    @with_response(
        200,
        schema=ProjectDataGetResponseSchema,
        description="Details of the project data",
    )
    def get(self, data_id):
        """
        Return the details of the project's data.

        The schema used to serialize the data will depend on the data_type
        attribute.

        :param int data_id: The id the project's data.
        :reqheader Authorization:
            JSON Web Token with Bearer scheme
            (``Authorization: Bearer <token>``).
        :status 403: The user is not allowed to access the project.
        :status 404: The data does not exist.

        .. :quickref: Project data; Get the details of the project's data
        """
        data = project_data_service.get_if_authorized(data_id)
        return dump_data(ProjectDataSchema(), project_data=data)

    @token_required
    @with_response(
        200,
        schema=MessageSchema,
        description="Project data successfully deleted",
    )
    def delete(self, data_id):
        """
        Delete the project's data.

        :param int data_id: The id the project's data.
        :status 403: The user is not allowed to access the project.
        :status 404: The data does not exist.

        .. :quickref: Project data; Delete the project's data
        """
        project_data = project_data_service.get_if_authorized(data_id)
        project_data.delete()
        return jsonify(message="Data {} deleted".format(data_id))

    @token_required
    @with_response(
        status=200,
        schema=ProjectDataGetResponseSchema,
        description="Updated details of the project data",
    )
    @with_request_body(
        ProjectDataSchema, description="New project data details"
    )
    def put(self, data_id):
        """
        Modify the project's data.

        :param int data_id: The id the project's data.
        :reqheader Authorization:
            JSON Web Token with Bearer scheme
            (``Authorization: Bearer <token>``).
        :status 400: There is an error in the json data.
        :status 403: The user is not allowed to access the project.
        :status 404: The data does not exist.

        .. :quickref: Project data; Modify the project's data
        """
        project_data = project_data_service.get_if_authorized(data_id)
        content = get_json_content()
        ps = ProjectDataProcessingService()
        input_data = content.pop("input_data", None)
        output_data = content.pop("output_data", None)
        if input_data is not None or output_data is not None:
            data_graph = ps.get_project_data_graph(project_data.project)
            new_output_data = []
            if input_data is not None:
                data_graph[project_data] = [
                    project_data_service.get_if_authorized(pdata_id)
                    for pdata_id in input_data
                ]
            if output_data is not None:
                deleted_outputs = []
                for pdata in project_data.output_data:
                    if pdata.id not in output_data:
                        deleted_outputs.append(pdata.id)
                        data_graph[pdata].remove(project_data)
                for pdata_id in output_data:
                    pdata = project_data_service.get_if_authorized(pdata_id)
                    if pdata not in project_data.output_data:
                        data_graph[pdata].append(project_data)
                    new_output_data.append(pdata)

            # Need to make graph with ids otherwise topological sort will
            # create new ProjectData when cloning the dictionary
            graph = {
                p.id: [idata.id for idata in inputs]
                for p, inputs in data_graph.items()
            }
            try:
                ps._topological_sort(graph)
            except TypeError:
                raise RequestError(
                    (
                        "change of data inputs create cycles in project data "
                        "graph"
                    ),
                    400,
                )

            if input_data is not None:
                prev_inputs = project_data.input_data
                project_data.input_data = []
                try:
                    for pdata in data_graph[project_data]:
                        project_data.add_input(pdata)
                except (TypeError, ValueError) as exc:
                    raise RequestError(str(exc), 400)

                # Detach cleanly old input data
                ps.detach_input(
                    project_data,
                    set(prev_inputs) - set(project_data.input_data),
                )
            if output_data is not None:
                prev_outputs = project_data.output_data
                project_data.output_data = []
                try:
                    for pdata in new_output_data:
                        pdata.add_input(project_data)
                except (TypeError, ValueError) as exc:
                    raise RequestError(str(exc), 400)

                for pdata in set(prev_outputs) | set(project_data.output_data):
                    if pdata not in project_data.output_data:
                        # Detach cleanly old output data
                        ps.detach_input(pdata, [project_data])
                    pdata.on_modification()
                    db.session.add(pdata)

        new_name = content.pop("name", project_data.name)
        if new_name != project_data.name:
            ps.rename_data(project_data, new_name)

        sch = ProjectDataSchema.type_schemas[project_data.data_type.name]()
        try:
            with db.session.no_autoflush:
                data = sch.load(content, instance=project_data)
        except NameError as error:
            raise RequestError(str(error), 400)
        if len(content) > 0 or input_data is not None:
            data.on_modification()
        db.session.commit()
        return dump_data(sch, project_data=data)


class ProjectDataActiveModelAPI(Resource):
    @token_required
    @with_response(
        status=200,
        schema=ProjectDataActiveModelPutResponseSchema,
        description="Updated active model of project data",
    )
    @with_request_body(
        schema=ProcessingModelSchema, description="New active model details"
    )
    def put(self, data_id):
        """
        Update the active processing model of the data.

        The type of processing model cannot be modified.

        :param int data_id: The id the project's data.
        :reqheader Authorization:
            JSON Web Token with Bearer scheme
            (``Authorization: Bearer <token>``).
        :status 403: The user is not allowed to access the project.
        :status 404: The data does not exist.

        .. :quickref: Project data; Modify the active processing model of the project data
        """  # noqa: E501
        project_data = project_data_service.get_if_authorized(data_id)
        content = get_json_content()
        model = project_data.processing_model
        schema = ProcessingModelSchema.type_schemas[model.model_type.name]
        schema = schema()
        model = schema.load(content, instance=model)
        model.update()
        project_data.update()
        return dump_data(schema, processing_model=model)


class ProjectDataModelAPI(Resource):
    @token_required
    @with_response(
        status=200,
        schema=ProjectDataActiveModelPutResponseSchema,
        description="Updated model of project data",
    )
    @with_request_body(
        schema=ProcessingModelSchema, description="New model details"
    )
    def put(self, data_id, model_id):
        """
        Update the processing model of the data.

        The type of processing model cannot be modified.

        :param int data_id: The id the project's data.
        :param int model_id: The id the processing model.
        :reqheader Authorization:
            JSON Web Token with Bearer scheme
            (``Authorization: Bearer <token>``).
        :status 403: The user is not allowed to access the project.
        :status 404:
            Either the project data or processing model does not exist.

        .. :quickref: Project data; Modify the processing model of the project data
        """  # noqa: E501
        project_data = project_data_service.get_if_authorized(data_id)
        models_list = project_data._processing_models
        for model in models_list:
            if model.id == model_id:
                break
        else:
            raise RequestError(
                "The processing model does not exist in this project data", 404
            )
        content = get_json_content()
        schema = ProcessingModelSchema.type_schemas[model.model_type.name]
        schema = schema()
        model = schema.load(content, instance=model)
        model.update()
        project_data.update()
        return dump_data(schema, processing_model=model)

    @token_required
    @with_response(
        200,
        schema=MessageSchema,
        description="Preference model deleted successfully",
    )
    def delete(self, data_id, model_id):
        """
        Delete the processing model of the data.

        :param int data_id: The id the project's data.
        :param int model_id: The id the processing model.
        :reqheader Authorization:
            JSON Web Token with Bearer scheme
            (``Authorization: Bearer <token>``).
        :status 400:
            The processing model is the only one for the data so cannot be
            deleted.
        :status 403: The user is not allowed to access the project.
        :status 404:
            Either the project data or processing model does not exist.

        .. :quickref: Project data; Delete the processing model of the project data
        """  # noqa: E501
        project_data = project_data_service.get_if_authorized(data_id)
        models_list = project_data._processing_models
        for i, model in enumerate(models_list):
            if model.id == model_id:
                break
        else:
            raise RequestError(
                "The processing model does not exist in this project data", 404
            )
        if len(models_list) == 1:
            raise RequestError(
                "The model cannot be deleted as it is the only model in the "
                "project data"
            )
        if model is project_data.processing_model:
            if i < len(models_list) - 1:
                project_data.processing_model = models_list[i + 1]
            else:
                project_data.processing_model = models_list[i - 1]
        model.delete()
        return jsonify(message="Model {} deleted".format(model_id))


class ProjectDataModelChangeAPI(Resource):
    @token_required
    @with_response(
        status=200,
        schema=ProjectDataActiveModelPutResponseSchema,
        description="Details of the project data active model",
    )
    @with_request_body(
        ProjectDataModelChangePostRequestBodySchema,
        description="Id of the processing model to set as the active model",
    )
    def post(self, data_id):
        """Change the active processing model.

        :param int data_id: The id of the project's data.
        :resjson int id: Id of the processing model to set as the active model.
        :reqheader Authorization:
            JSON Web Token with Bearer scheme
            (``Authorization: Bearer <token>``).
        :status 403: The user is not allower to access the project.
        :status 404: The project data or processing model does not exist.
        :status 400: The content is incorrect.

        .. :quickref: Project data; Change the processing model of the project data.
        """  # noqa: E501
        project_data = project_data_service.get_if_authorized(data_id)
        content = get_json_content()
        try:
            model_id = content.pop("id")
        except KeyError:
            raise RequestError("No processing_model id found")
        if len(content) > 0:
            raise RequestError(
                "The post data should only contain the model id"
            )
        for model in project_data._processing_models:
            if model.id == model_id:
                break
        else:
            raise RequestError(
                "The processing model does not exist for this project data",
                404,
            )
        project_data._processing_model = model
        project_data.update()
        schema = ProcessingModelSchema.type_schemas[model.model_type.name]
        schema = schema()
        return dump_data(schema, processing_model=model)


class ProjectDataInputSubgraphAPI(Resource):
    @token_required
    @with_response(
        200,
        schema=ProjectSharedDataListGetResponseSchema,
        description="Project data input subgraph",
    )
    def get(self, data_id):
        """
        Return the input data subgraph spanning from the project data.

        :param int data_id: The id the project's data.
        :reqheader Authorization:
            JSON Web Token with Bearer scheme
            (``Authorization: Bearer <token>``).
        :status 403: The user is not allowed to access the project data.
        :status 404: The project data doesn't exist.

        .. :quickref: Project data; Get the project data input subgraph
        """
        project_data = project_data_service.get_if_authorized(data_id)
        ps = ProjectDataProcessingService()
        graph = ps.get_input_graph(project_data)

        dump_schemas = {
            DataType.geo_data: DataGeoSchema(
                only=(
                    "name",
                    "id",
                    "data",
                    "data_type",
                    "is_outdated",
                    "input_data",
                    "output_data",
                )
            ),
            DataType.generator: DataGeneratorSchema(
                only=(
                    "name",
                    "id",
                    "data",
                    "data_type",
                    "is_outdated",
                    "processing_model",
                    "input_data",
                    "output_data",
                )
            ),
            DataType.global_data: ProjectGlobalDataSchema(
                only=(
                    "name",
                    "id",
                    "data",
                    "data_type",
                    "is_outdated",
                    "input_data",
                    "output_data",
                )
            ),
            DataType.data_stream: DataStreamSchema(
                only=(
                    "name",
                    "id",
                    "data",
                    "data_type",
                    "is_outdated",
                    "input_data",
                    "output_data",
                    "stream",
                )
            ),
        }
        res = []
        for projdata in graph:
            res.append(dump_schemas[projdata.data_type].dump(projdata))
        return {"data_list": res}


class ProjectDataOutputSubgraphAPI(Resource):
    @token_required
    @with_response(
        200,
        schema=ProjectSharedDataListGetResponseSchema,
        description="Project data output subgraph",
    )
    def get(self, data_id):
        """
        Return the output data subgraph spanning from the project data.

        :param int data_id: The id the project's data.
        :reqheader Authorization:
            JSON Web Token with Bearer scheme
            (``Authorization: Bearer <token>``).
        :status 403: The user is not allowed to access the project data.
        :status 404: The project data doesn't exist.

        .. :quickref: Project data; Get the project data output subgraph
        """
        project_data = project_data_service.get_if_authorized(data_id)
        ps = ProjectDataProcessingService()
        graph = ps.get_output_graph(project_data)

        dump_schemas = {
            DataType.geo_data: DataGeoSchema(
                only=(
                    "name",
                    "id",
                    "data",
                    "data_type",
                    "is_outdated",
                    "input_data",
                    "output_data",
                )
            ),
            DataType.generator: DataGeneratorSchema(
                only=(
                    "name",
                    "id",
                    "data",
                    "data_type",
                    "is_outdated",
                    "processing_model",
                    "input_data",
                    "output_data",
                )
            ),
            DataType.global_data: ProjectGlobalDataSchema(
                only=(
                    "name",
                    "id",
                    "data",
                    "data_type",
                    "is_outdated",
                    "input_data",
                    "output_data",
                )
            ),
            DataType.data_stream: DataStreamSchema(
                only=(
                    "name",
                    "id",
                    "data",
                    "data_type",
                    "is_outdated",
                    "input_data",
                    "output_data",
                    "stream",
                )
            ),
        }
        res = []
        for projdata in graph:
            res.append(dump_schemas[projdata.data_type].dump(projdata))
        return {"data_list": res}


class ProjectDataReplaceInputAttributesAPI(Resource):
    @token_required
    @with_response(
        200,
        schema=ProjectDataGetResponseSchema,
        description="Details of the project data",
    )
    @with_request_body(schema=ProjectDataReplaceInputsPostRequestBodySchema)
    def post(self, data_id):
        """Replace project data input attributes.

        It replaces usage of old attributes given by new attributes given,
        assuming the old attributes given are attributes of this project's data
        and that the new attributes given are not already used as inputs
        alongside some of the old attributes.

        Its main usage is threfore to replace whole project data
        (all their attributes) by completely unused other project data.

        :param int data_id: The id the project's data.
        :reqheader Authorization:
            JSON Web Token with Bearer scheme
            (``Authorization: Bearer <token>``).
        :status 400: Remapping could not proceed with given arguments.
        :status 403: The user is not allowed to access the project.
        :status 404: The data does not exist.

        .. :quickref: Project data; Replace project data input attributes
        """
        data = project_data_service.get_if_authorized(data_id)
        content = get_json_content()
        args = ProjectDataReplaceInputsPostRequestBodySchema().load(content)

        attributes_mapping = {}
        for item in args["attributes_mapping"]:
            old_attr_id = item["old_attribute"]
            new_attr_id = item["new_attribute"]
            old_attr = DataAttribute.get_by_id(old_attr_id)
            new_attr = DataAttribute.get_by_id(new_attr_id)
            if old_attr is None or new_attr is None:
                raise RequestError("attributes mapping is invalid", 400)
            attributes_mapping[old_attr] = new_attr
        try:
            ProjectDataProcessingService().replace_input_attributes(
                data, attributes_mapping
            )
        except (ValueError, TypeError) as exc:
            raise RequestError(str(exc), 400)

        ps = ProjectDataProcessingService()
        data_graph = ps.get_project_data_graph(data.project)
        # Need to make graph with ids otherwise topological sort will create
        # new ProjectData when cloning the dictionary
        graph = {
            p.id: [idata.id for idata in inputs]
            for p, inputs in data_graph.items()
        }
        try:
            ps._topological_sort(graph)
        except TypeError:
            raise RequestError(
                "new data input creates cycles in project data graph", 400
            )
        data.save()

        return dump_data(ProjectDataSchema(), project_data=data)


class MRSortInferenceAPI(Resource):
    @token_required
    @with_response(
        200,
        schema=MRSortInferenceGetResponseSchema,
        description="MR-Sort inference model details",
    )
    def get(self, data_id):
        """
        Return the details of the MR-Sort inference model for the project's
        data.

        :param int data_id: The id the project's data.
        :reqheader Authorization:
            JSON Web Token with Bearer scheme
            (``Authorization: Bearer <token>``).
        :status 400: The processing model is not an MR-Sort model.
        :status 403: The user is not allowed to access the project.
        :status 404: The data does not exist.

        .. :quickref: MR-Sort; Get the MR-Sort inference model details
        """
        project_data = project_data_service.get_if_authorized(data_id)
        model = project_data.processing_model
        if not isinstance(project_data.processing_model, MRSort):
            raise RequestError("Only works with an MR-Sort model")
        return dump_data(MRSortInferenceSchema(), inference_data=model)

    def infer_mrsort(self, model):
        try:
            processing_model_service.infer_model(model)
        except InferenceError:
            raise RequestError("Error in the inference of the model", 500)
        return "ok"

    def load_alternatives(self, model):
        processing_model_service.load_mrsort_inference_data(model)
        return dump_data(MRSortInferenceSchema(), inference_data=model)

    @token_required
    @with_response(
        200,
        schema=MRSortInferenceGetResponseSchema,
        content={
            "text/plain": {"schema": {"type": "string", "example": "ok"}}
        },
        description=(
            "MR-Sort inference model details if action ``load`` is successful,"
            " 'ok' if action ``infer`` is successful"
        ),
    )
    @with_query_arg(
        "action",
        description=(
            "The action to execute:\n- ``load`` will load the inference "
            "alternatives\n- ``infer`` will try to infer the MR-Sort model "
            "from the alternatives"
        ),
        schema={"type": "string", "enum": ["infer", "load"]},
    )
    def post(self, data_id):
        """
        Execute an action on the MR-Sort inference model.

        :param int data_id: The id the project's data.
        :query str action:
            The action to execute: ``load`` will load the inference
            alternatives, ``infer`` will try to infer the MR-Sort model from
            the alternatives (**required**).
        :reqheader Authorization:
            JSON Web Token with Bearer scheme
            (``Authorization: Bearer <token>``).
        :status 400:
            The model is not an MR-Sort model or ``action`` is incorrect.
        :status 403: The user is not allowed to access the project.
        :status 404: The data does not exist.
        :status 500: Inference error

        .. :quickref: MR-Sort; Execute an action on the MR-Sort inference model
        """
        parser = reqparse.RequestParser()
        parser.add_argument("action", type=str, location="args", required=True)
        args = parser.parse_args()
        project_data = project_data_service.get_if_authorized(data_id)
        model = project_data.processing_model
        if not isinstance(project_data.processing_model, MRSort):
            raise RequestError("Only works with an MR-Sort model")
        if args.get("action") == "load":
            return self.load_alternatives(model)
        elif args.get("action") == "infer":
            return self.infer_mrsort(model)
        else:
            raise RequestError("Unknwon action")

    @token_required
    @with_response(
        200,
        schema=MRSortInferenceGetResponseSchema,
        description="New MR-Sort inference model details",
    )
    @with_request_body(
        schema=MRSortInferenceSchema,
        description="New MR-Sort inference model details",
    )
    def put(self, data_id):
        """
        Modify the MR-Sort inference model.

        :param int data_id: The id the project's data.
        :reqheader Authorization:
            JSON Web Token with Bearer scheme
            (``Authorization: Bearer <token>``).
        :status 400: The processing model is not an MR-Sort model.
        :status 403: The user is not allowed to access the project.
        :status 404: The data does not exist.

        .. :quickref: MR-Sort; Modify the MR-Sort inference model
        """
        project_data = project_data_service.get_if_authorized(data_id)
        model = project_data.processing_model
        if not isinstance(model, MRSort):
            raise RequestError("Only works with an MR-Sort model")
        content = get_json_content()
        schema = MRSortInferenceSchema()
        model = schema.load(content, instance=model)
        model.update()
        project_data.update()
        return dump_data(schema, inference_data=model)



"""This module contains all schemas used as response/requestBody types.
"""

from marshmallow import Schema, fields

from ..models import (
    DataType,
    GeoDataType,
    ProcessingModel,
    ProjectData,
    ProjectTaskType,
)
from .base import FieldPluck, OneOfSchema, constant_enum
from .geo_data import (
    BaseGeoDataSchema,
    DataAttributeSchema,
    DataAttributeSchemaWithValues,
    FeatureSchema,
    GeoDataAccessSchema,
    GeoDataSchema,
    GeoOnlyDataSchema,
    GlobalDataAccessSchema,
    StreamGeoDataAccessSchema,
    StreamGeoDataSchema,
)
from .mrsort_inference import MRSortInferenceSchema
from .processing_models import ProcessingModelSchema
from .project import (
    ProjectAccessSchema,
    ProjectSchema,
    TemplateAccessSchema,
    TemplateSchema,
)
from .project_data import (
    DataStreamSubStreamSchema,
    ProjectDataBaseSchema,
    ProjectDataSchema,
)
from .schemas import ProjectTaskSchema, UserSchema
from .share import DataShareSchema


class MessageSchema(Schema):
    """This represents a simple message.

    Mainly used for error responses.
    """

    message = fields.String()


class DataAttributeGetResponse(Schema):
    attribute = fields.Nested(DataAttributeSchema)


class DataAttributeWithValuesGetResponse(Schema):
    attribute = fields.Nested(DataAttributeSchemaWithValues)


class GeoDataGetResponse(Schema):
    geodata = fields.Nested(GeoOnlyDataSchema())


class GeoDataListResponseSchema(Schema):
    geodata = fields.Nested(GeoDataSchema, many=True)


class GeoDataAccessResponseSchema(Schema):
    access = fields.Nested(GeoDataAccessSchema)


class StreamGeoDataResponseSchema(Schema):
    stream = fields.Nested(StreamGeoDataSchema)


class StreamGeoDataListResponseSchema(Schema):
    stream = fields.Nested(StreamGeoDataSchema, many=True)


class StreamGeoDataAccessResponseSchema(Schema):
    access = fields.Nested(StreamGeoDataAccessSchema)


class StreamGeoDataPostRequestSchema(Schema):
    url = fields.String(required=True)


class WMSGeoDataPostRequestSchema(StreamGeoDataPostRequestSchema):
    classes = fields.List(fields.List(fields.Float), required=False)


class WMSStreamClassesSchema(Schema):
    classes = fields.List(fields.List(fields.Float), required=False)
    step = fields.Float(required=False)
    start = fields.Float(required=False)
    stop = fields.Float(required=False)
    resolution = fields.Float(required=False)


class FeatureResponseSchema(Schema):
    feature = fields.Nested(FeatureSchema)


class UserAccessTokenResponseSchema(Schema):
    access_token = fields.String()
    token_type = fields.String()


class UserTokenResponseSchema(UserAccessTokenResponseSchema):
    refresh_token = fields.String()


class ProjectDataListGetResponseSchema(Schema):
    project_data = fields.Nested(ProjectDataSchema, many=True)


class ProjectProcessTaskSchema(ProjectTaskSchema):
    type = fields.Enum(constant_enum(ProjectTaskType.process_project_data))


class ProjectDataPostResponseSchema(Schema):
    task = fields.Nested(ProjectProcessTaskSchema)


class ProjectDataGetResponseSchema(Schema):
    project_data = fields.Nested(ProjectDataSchema)


class ProjectDataInputListPostResponseSchema(Schema):
    input_data = fields.Nested(
        ProjectDataBaseSchema, only=("id", "name", "data_type"), many=True
    )


class ProjectDataActiveModelPutResponseSchema(Schema):
    processing_model = fields.Nested(ProcessingModelSchema)


class ProjectDataModelChangePostRequestBodySchema(Schema):
    id = FieldPluck(ProcessingModel, "id", required=True)


class MRSortInferenceGetResponseSchema(Schema):
    inference_data = fields.Nested(MRSortInferenceSchema)


class ProjectListGetResponseSchema(Schema):
    projects = fields.Nested(ProjectSchema, many=True)


class ProjectFromTemplateRequestBody(Schema):
    template = fields.Integer(required=True)
    name = fields.String(allow_none=True)
    description = fields.String(allow_none=True)
    extent = fields.List(
        fields.Float(),
        metadata={"minItems": 4, "maxItems": 4},
        allow_none=True,
    )


class ProjectGetResponseSchema(Schema):
    project = fields.Nested(ProjectSchema)


class ProjectPermissionGetResponseSchema(Schema):
    access = fields.Nested(ProjectAccessSchema)


class TemplatePermissionGetResponseSchema(Schema):
    access = fields.Nested(TemplateAccessSchema)


class ProjectSharedDataListPostResponseSchema(Schema):
    data_list = fields.Nested(
        ProjectDataBaseSchema,
        only=("name", "id", "data", "data_type"),
        many=True,
    )


class ProjectSharedBaseDataSchema(Schema):
    id = fields.Integer()
    name = fields.String()
    data = fields.Pluck(BaseGeoDataSchema, "id")
    data_type = fields.Enum(DataType)
    is_outdated = fields.Boolean()
    input_data = FieldPluck(ProjectData, "id", many=True)
    output_data = FieldPluck(ProjectData, "id", many=True)


class ProjectSharedGeoDataSchema(ProjectSharedBaseDataSchema):
    data_type = fields.Enum(constant_enum(DataType.geo_data))


class ProjectSharedGlobalDataSchema(ProjectSharedBaseDataSchema):
    data_type = fields.Enum(constant_enum(DataType.global_data))


class ProjectSharedDataStreamSchema(ProjectSharedBaseDataSchema):
    data_type = fields.Enum(constant_enum(DataType.data_stream))
    stream = fields.Nested(
        DataStreamSubStreamSchema,
        only=("id", "type"),
    )


class ProjectSharedGeneratedDataSchema(ProjectSharedBaseDataSchema):
    data_type = fields.Enum(constant_enum(DataType.generator))
    processing_model = fields.Nested(
        ProcessingModelSchema(exclude=("data_generator",))
    )


class ProjectSharedDataSchema(OneOfSchema, ProjectSharedBaseDataSchema):
    type_field = "data_type"
    type_schemas = {
        DataType.geo_data.name: ProjectSharedGeoDataSchema,
        DataType.generator.name: ProjectSharedGeneratedDataSchema,
        DataType.global_data.name: ProjectSharedGlobalDataSchema,
        DataType.data_stream.name: ProjectSharedDataStreamSchema,
    }

    def get_obj_type(self, obj):
        return obj.data_type.name


class ProjectSharedDataListGetResponseSchema(Schema):
    data_list = fields.Nested(ProjectSharedDataSchema, many=True)


class ProjectTemplatePostResponseSchema(Schema):
    template = fields.Nested(TemplateSchema)


class ProjectTemplateListGetResponseSchema(Schema):
    templates = fields.Nested(TemplateSchema, many=True)


class ProjectSharesGetResponseSchema(Schema):
    shared = fields.Nested(DataShareSchema, many=True)


class ShareBaseDataPostResponseSchema(Schema):
    shared = fields.Nested(DataShareSchema)


class ProjectTaskListGetResponseSchema(Schema):
    tasks = fields.Nested(ProjectTaskSchema, many=True)
    total = fields.Integer()


class TaskGetResponseSchema(Schema):
    task = fields.Nested(ProjectTaskSchema)


class UserListGetResponseSchema(Schema):
    users = fields.Nested(UserSchema, only=("id", "username"), many=True)


class GlobalDataAccessResponseSchema(Schema):
    access = fields.Nested(GlobalDataAccessSchema)


class DataAttributeRemappingSchema(Schema):
    old_attribute = fields.Integer(required=True, allow_none=False)
    new_attribute = fields.Integer(required=True, allow_none=False)


class ProjectDataReplaceInputsPostRequestBodySchema(Schema):
    attributes_mapping = fields.Nested(
        DataAttributeRemappingSchema,
        many=True,
        required=True,
        allow_none=False,
        metadata={
            "description": "list mapping of old data attributes to new ones"
        },
    )


class DataTypeSetSchema(Schema):
    type = fields.List(fields.Enum(GeoDataType))



from functools import wraps

from flask import request

from ..exceptions import PermissionError
from ..models import User, db
from .auth_service import check_jwt_authentication


def has_permission(obj, user=None):
    """Check if the user of the request is authorized to use the object

    Returns a boolean indicating if the user has permissions on the object.

    Args:
        user:
            The user to check for the permission.
            Default: None (load the user from the token).
    """
    if user is None:
        user = check_jwt_authentication(request)
    if user is None:
        return False
    return obj.is_user_authorized(user)


def has_permission_for_id(obj_class, obj_id):
    """Check if the user of the request is authorized to use the object with
    the given id"""
    user = check_jwt_authentication(request)
    if user is None:
        return False
    obj = obj_class.get_by_id(obj_id)
    return obj.is_user_authorized(user)


def has_ability(ability):
    """Check that a user has the necessary ability

    Executes the decorated function if the user has the abilty, otherwise
    raises an error.

    Raises:
        PermissionError if the user does not have the ability
    """

    def decorator(fn):
        @wraps(fn)
        def wrapper(*args, **kwargs):
            user = check_jwt_authentication(request)
            if any([ability is p.ability for p in user.permissions]):
                return fn(*args, **kwargs)
            raise PermissionError

        return wrapper

    return decorator


def get_all_users():
    """Returns the list of all the users in the application"""
    return db.session.query(User).order_by(User.id).all()



"""Initialize database

Revision ID: aa42863f0b44
Revises:
Create Date: 2024-09-19 08:18:04.298905

"""

import sqlalchemy as sa
from alembic import op
from geoalchemy2 import Geometry
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = "aa42863f0b44"
down_revision = None
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_geospatial_table(
        "base_geo_data",
        sa.Column(
            "type",
            sa.Enum(
                "geo_data",
                "generated_geo_data",
                "proposition_geo_data",
                name="geodatatype",
            ),
            nullable=True,
        ),
        sa.Column("name", sa.String(), nullable=False),
        sa.Column(
            "extent",
            Geometry(
                spatial_index=False,
                from_text="ST_GeomFromEWKT",
                name="geometry",
            ),
            nullable=True,
        ),
        sa.Column("properties_modified_at", sa.DateTime(), nullable=True),
        sa.Column("created_at", sa.DateTime(), nullable=True),
        sa.Column("modified_at", sa.DateTime(), nullable=True),
        sa.Column("id", sa.Integer(), nullable=False),
        sa.PrimaryKeyConstraint("id", name=op.f("pk_base_geo_data")),
    )
    with op.batch_alter_table("base_geo_data", schema=None) as batch_op:
        batch_op.create_geospatial_index(
            "idx_base_geo_data_extent",
            ["extent"],
            unique=False,
            postgresql_using="gist",
            postgresql_ops={},
        )

    op.create_table(
        "preference_model",
        sa.Column(
            "pref_type",
            sa.Enum(
                "categories_rule",
                "continuous_rule",
                "geo_buffer",
                "mrsort",
                "weighted_sum",
                name="preftype",
            ),
            nullable=True,
        ),
        sa.Column("name", sa.String(), nullable=True),
        sa.Column("data_generator_id", sa.Integer(), nullable=False),
        sa.Column(
            "keep_overlap",
            sa.Enum("all", "min", "max", "sum", "average", name="keepoverlap"),
            nullable=False,
        ),
        sa.Column("cut_to_extent", sa.Boolean(), nullable=True),
        sa.Column("dissolve_adjacent", sa.Boolean(), nullable=True),
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("radius", sa.BigInteger(), nullable=True),
        sa.Column("majority_threshold", sa.Float(), nullable=True),
        sa.Column("rule", sa.String(), nullable=True),
        sa.PrimaryKeyConstraint("id", name=op.f("pk_preference_model")),
    )
    op.create_table(
        "project_data",
        sa.Column(
            "data_type",
            sa.Enum("geo_data", "generator", name="datatype"),
            nullable=True,
        ),
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("data_id", sa.Integer(), nullable=True),
        sa.Column("last_update", sa.DateTime(), nullable=True),
        sa.Column("project_id", sa.Integer(), nullable=False),
        sa.Column("description", sa.String(), nullable=True),
        sa.Column("name", sa.String(), nullable=True),
        sa.Column("active_model_id", sa.Integer(), nullable=True),
        sa.ForeignKeyConstraint(
            ["active_model_id"],
            ["preference_model.id"],
            name=op.f("fk_project_data_active_model_id_preference_model"),
        ),
        sa.ForeignKeyConstraint(
            ["data_id"],
            ["base_geo_data.id"],
            name=op.f("fk_project_data_data_id_base_geo_data"),
        ),
        sa.PrimaryKeyConstraint("id", name=op.f("pk_project_data")),
    )
    op.create_foreign_key(
        op.f("fk_preference_model_data_generator_id_project_data"),
        "preference_model",
        "project_data",
        ["data_generator_id"],
        ["id"],
    )
    op.create_table(
        "user",
        sa.Column("username", sa.String(), nullable=True),
        sa.Column("email", sa.String(), nullable=True),
        sa.Column("password_hash", sa.String(), nullable=True),
        sa.Column("id", sa.Integer(), nullable=False),
        sa.PrimaryKeyConstraint("id", name=op.f("pk_user")),
        sa.UniqueConstraint("email", name=op.f("uq_user_email")),
        sa.UniqueConstraint("username", name=op.f("uq_user_username")),
    )
    op.create_table(
        "data_attribute",
        sa.Column(
            "type",
            sa.Enum(
                "quantitative", "nominal", "ordinal", name="attributetype"
            ),
            nullable=True,
        ),
        sa.Column("name", sa.String(), nullable=False),
        sa.Column(
            "statistics",
            postgresql.JSONB(astext_type=sa.Text()),
            nullable=True,
        ),
        sa.Column("geo_data_id", sa.Integer(), nullable=False),
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column(
            "order", postgresql.JSONB(astext_type=sa.Text()), nullable=True
        ),
        sa.ForeignKeyConstraint(
            ["geo_data_id"],
            ["base_geo_data.id"],
            name=op.f("fk_data_attribute_geo_data_id_base_geo_data"),
        ),
        sa.PrimaryKeyConstraint("id", name=op.f("pk_data_attribute")),
    )
    with op.batch_alter_table("data_attribute", schema=None) as batch_op:
        batch_op.create_index(
            batch_op.f("ix_data_attribute_geo_data_id"),
            ["geo_data_id"],
            unique=False,
        )

    op.create_table(
        "data_input_association",
        sa.Column("project_data_id", sa.Integer(), nullable=False),
        sa.Column("input_data_id", sa.Integer(), nullable=False),
        sa.ForeignKeyConstraint(
            ["input_data_id"],
            ["project_data.id"],
            name=op.f("fk_data_input_association_input_data_id_project_data"),
        ),
        sa.ForeignKeyConstraint(
            ["project_data_id"],
            ["project_data.id"],
            name=op.f(
                "fk_data_input_association_project_data_id_project_data"
            ),
        ),
        sa.PrimaryKeyConstraint(
            "project_data_id",
            "input_data_id",
            name=op.f("pk_data_input_association"),
        ),
    )
    op.create_table(
        "data_share",
        sa.Column("data_id", sa.Integer(), nullable=False),
        sa.Column("uid", sa.UUID(), nullable=False),
        sa.Column("expiration", sa.DateTime(), nullable=True),
        sa.Column("expired", sa.Boolean(), nullable=True),
        sa.Column("id", sa.Integer(), nullable=False),
        sa.ForeignKeyConstraint(
            ["data_id"],
            ["project_data.id"],
            name=op.f("fk_data_share_data_id_project_data"),
        ),
        sa.PrimaryKeyConstraint("id", name=op.f("pk_data_share")),
        sa.UniqueConstraint("uid", name=op.f("uq_data_share_uid")),
    )
    op.create_table(
        "discrete_category",
        sa.Column("type", sa.String(), nullable=True),
        sa.Column("name", sa.String(), nullable=True),
        sa.Column("position", sa.Integer(), nullable=True),
        sa.Column("preference_model_id", sa.Integer(), nullable=True),
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column(
            "rules", postgresql.ARRAY(sa.String(), dimensions=1), nullable=True
        ),
        sa.ForeignKeyConstraint(
            ["preference_model_id"],
            ["preference_model.id"],
            name=op.f(
                "fk_discrete_category_preference_model_id_preference_model"
            ),
            ondelete="CASCADE",
        ),
        sa.PrimaryKeyConstraint("id", name=op.f("pk_discrete_category")),
    )
    op.create_geospatial_table(
        "feature",
        sa.Column(
            "geom",
            Geometry(
                spatial_index=False,
                from_text="ST_GeomFromEWKT",
                name="geometry",
            ),
            nullable=True,
        ),
        sa.Column("geo_data_id", sa.Integer(), nullable=True),
        sa.Column(
            "explainability",
            postgresql.JSONB(astext_type=sa.Text()),
            nullable=True,
        ),
        sa.Column("id", sa.Integer(), nullable=False),
        sa.ForeignKeyConstraint(
            ["geo_data_id"],
            ["base_geo_data.id"],
            name=op.f("fk_feature_geo_data_id_base_geo_data"),
        ),
        sa.PrimaryKeyConstraint("id", name=op.f("pk_feature")),
    )
    with op.batch_alter_table("feature", schema=None) as batch_op:
        batch_op.create_geospatial_index(
            "idx_feature_geom",
            ["geom"],
            unique=False,
            postgresql_using="gist",
            postgresql_ops={},
        )
        batch_op.create_index(
            batch_op.f("ix_feature_geo_data_id"), ["geo_data_id"], unique=False
        )

    op.create_geospatial_table(
        "project",
        sa.Column("is_template", sa.Boolean(), nullable=False),
        sa.Column("name", sa.String(), nullable=False),
        sa.Column("manager_id", sa.Integer(), nullable=True),
        sa.Column("description", sa.String(), nullable=True),
        sa.Column(
            "extent",
            Geometry(
                spatial_index=False,
                from_text="ST_GeomFromEWKT",
                name="geometry",
                nullable=False,
            ),
            nullable=False,
        ),
        sa.Column("created_at", sa.DateTime(), nullable=True),
        sa.Column("modified_at", sa.DateTime(), nullable=True),
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("is_public", sa.Boolean(), nullable=False),
        sa.ForeignKeyConstraint(
            ["manager_id"],
            ["user.id"],
            name=op.f("fk_project_manager_id_user"),
        ),
        sa.PrimaryKeyConstraint("id", name=op.f("pk_project")),
    )
    with op.batch_alter_table("project", schema=None) as batch_op:
        batch_op.create_geospatial_index(
            "idx_project_extent",
            ["extent"],
            unique=False,
            postgresql_using="gist",
            postgresql_ops={},
        )
    op.create_foreign_key(
        op.f("fk_project_data_project_id_project"),
        "project_data",
        "project",
        ["project_id"],
        ["id"],
    )

    op.create_table(
        "stream_geo_data",
        sa.Column(
            "type", sa.Enum("wfs", "wms", name="streamdatatype"), nullable=True
        ),
        sa.Column("name", sa.String(), nullable=False),
        sa.Column("title", sa.String(), nullable=True),
        sa.Column("original_name", sa.String(), nullable=True),
        sa.Column("url", sa.String(), nullable=False),
        sa.Column("upload_user_id", sa.Integer(), nullable=True),
        sa.Column("description", sa.String(), nullable=True),
        sa.Column(
            "keywords",
            postgresql.ARRAY(sa.String(), dimensions=1),
            nullable=True,
        ),
        sa.Column("is_public", sa.Boolean(), nullable=False),
        sa.Column("created_at", sa.DateTime(), nullable=True),
        sa.Column("modified_at", sa.DateTime(), nullable=True),
        sa.Column("id", sa.Integer(), nullable=False),
        sa.ForeignKeyConstraint(
            ["upload_user_id"],
            ["user.id"],
            name=op.f("fk_stream_geo_data_upload_user_id_user"),
        ),
        sa.PrimaryKeyConstraint("id", name=op.f("pk_stream_geo_data")),
    )
    op.create_table(
        "user_permission",
        sa.Column("user_id", sa.Integer(), nullable=False),
        sa.Column(
            "ability",
            sa.Enum(
                "create_project", "create_geo_data", name="permissionability"
            ),
            nullable=False,
        ),
        sa.Column("id", sa.Integer(), nullable=False),
        sa.ForeignKeyConstraint(
            ["user_id"],
            ["user.id"],
            name=op.f("fk_user_permission_user_id_user"),
        ),
        sa.PrimaryKeyConstraint("id", name=op.f("pk_user_permission")),
    )
    op.create_table(
        "user_refresh_token",
        sa.Column("user_id", sa.Integer(), nullable=True),
        sa.Column("refresh_token", sa.LargeBinary(), nullable=True),
        sa.Column("expiration", sa.DateTime(), nullable=True),
        sa.Column("id", sa.Integer(), nullable=False),
        sa.ForeignKeyConstraint(
            ["user_id"],
            ["user.id"],
            name=op.f("fk_user_refresh_token_user_id_user"),
        ),
        sa.PrimaryKeyConstraint("id", name=op.f("pk_user_refresh_token")),
    )
    op.create_table(
        "data_value",
        sa.Column(
            "type",
            sa.Enum(
                "quantitative", "nominal", "ordinal", name="attributetype"
            ),
            nullable=True,
        ),
        sa.Column("attribute_id", sa.Integer(), nullable=True),
        sa.Column("feature_id", sa.Integer(), nullable=True),
        sa.Column(
            "value", postgresql.JSONB(astext_type=sa.Text()), nullable=True
        ),
        sa.Column("id", sa.Integer(), nullable=False),
        sa.ForeignKeyConstraint(
            ["attribute_id"],
            ["data_attribute.id"],
            name=op.f("fk_data_value_attribute_id_data_attribute"),
        ),
        sa.ForeignKeyConstraint(
            ["feature_id"],
            ["feature.id"],
            name=op.f("fk_data_value_feature_id_feature"),
            ondelete="CASCADE",
        ),
        sa.PrimaryKeyConstraint("id", name=op.f("pk_data_value")),
        sa.UniqueConstraint(
            "attribute_id",
            "feature_id",
            name=op.f("uq_data_value_attribute_id"),
        ),
    )
    with op.batch_alter_table("data_value", schema=None) as batch_op:
        batch_op.create_index(
            batch_op.f("ix_data_value_attribute_id"),
            ["attribute_id"],
            unique=False,
        )
        batch_op.create_index(
            batch_op.f("ix_data_value_feature_id"),
            ["feature_id"],
            unique=False,
        )

    op.create_table(
        "default_value",
        sa.Column("attribute_id", sa.Integer(), nullable=False),
        sa.Column("model_id", sa.Integer(), nullable=False),
        sa.Column(
            "value", postgresql.JSONB(astext_type=sa.Text()), nullable=True
        ),
        sa.ForeignKeyConstraint(
            ["attribute_id"],
            ["data_attribute.id"],
            name=op.f("fk_default_value_attribute_id_data_attribute"),
            ondelete="CASCADE",
        ),
        sa.ForeignKeyConstraint(
            ["model_id"],
            ["preference_model.id"],
            name=op.f("fk_default_value_model_id_preference_model"),
            ondelete="CASCADE",
        ),
        sa.PrimaryKeyConstraint(
            "attribute_id", "model_id", name=op.f("pk_default_value")
        ),
    )
    op.create_table(
        "geo_data",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("original_name", sa.String(), nullable=True),
        sa.Column("source_driver", sa.String(), nullable=True),
        sa.Column("upload_user_id", sa.Integer(), nullable=True),
        sa.Column("description", sa.String(), nullable=True),
        sa.Column("stream_id", sa.Integer(), nullable=True),
        sa.Column("is_public", sa.Boolean(), nullable=False),
        sa.ForeignKeyConstraint(
            ["id"],
            ["base_geo_data.id"],
            name=op.f("fk_geo_data_id_base_geo_data"),
        ),
        sa.ForeignKeyConstraint(
            ["stream_id"],
            ["stream_geo_data.id"],
            name=op.f("fk_geo_data_stream_id_stream_geo_data"),
        ),
        sa.ForeignKeyConstraint(
            ["upload_user_id"],
            ["user.id"],
            name=op.f("fk_geo_data_upload_user_id_user"),
        ),
        sa.PrimaryKeyConstraint("id", name=op.f("pk_geo_data")),
    )
    op.create_table(
        "mrsort_criterion",
        sa.Column(
            "profiles",
            postgresql.ARRAY(sa.Float(), dimensions=1),
            nullable=True,
        ),
        sa.Column("weight", sa.Float(), nullable=True),
        sa.Column("maximize", sa.Boolean(), nullable=False),
        sa.Column("mrsort_id", sa.Integer(), nullable=False),
        sa.Column("attribute_id", sa.Integer(), nullable=False),
        sa.ForeignKeyConstraint(
            ["attribute_id"],
            ["data_attribute.id"],
            name=op.f("fk_mrsort_criterion_attribute_id_data_attribute"),
            ondelete="CASCADE",
        ),
        sa.ForeignKeyConstraint(
            ["mrsort_id"],
            ["preference_model.id"],
            name=op.f("fk_mrsort_criterion_mrsort_id_preference_model"),
            ondelete="CASCADE",
        ),
        sa.PrimaryKeyConstraint(
            "mrsort_id", "attribute_id", name=op.f("pk_mrsort_criterion")
        ),
    )
    op.create_table(
        "mrsort_inference_alternative",
        sa.Column("category_id", sa.Integer(), nullable=True),
        sa.Column("mrsort_id", sa.Integer(), nullable=False),
        sa.Column("id", sa.Integer(), nullable=False),
        sa.ForeignKeyConstraint(
            ["category_id"],
            ["discrete_category.id"],
            name=op.f(
                "fk_mrsort_inference_alternative_category_id_discrete_category"
            ),
            ondelete="SET NULL",
        ),
        sa.ForeignKeyConstraint(
            ["mrsort_id"],
            ["preference_model.id"],
            name=op.f(
                "fk_mrsort_inference_alternative_mrsort_id_preference_model"
            ),
            ondelete="CASCADE",
        ),
        sa.PrimaryKeyConstraint(
            "id", name=op.f("pk_mrsort_inference_alternative")
        ),
    )
    op.create_table(
        "project_permission",
        sa.Column("object_id", sa.Integer(), nullable=True),
        sa.Column("user_id", sa.Integer(), nullable=False),
        sa.Column("id", sa.Integer(), nullable=False),
        sa.ForeignKeyConstraint(
            ["object_id"],
            ["project.id"],
            name=op.f("fk_project_permission_object_id_project"),
        ),
        sa.ForeignKeyConstraint(
            ["user_id"],
            ["user.id"],
            name=op.f("fk_project_permission_user_id_user"),
        ),
        sa.PrimaryKeyConstraint("id", name=op.f("pk_project_permission")),
    )
    op.create_table(
        "project_root",
        sa.Column("project_id", sa.Integer(), nullable=False),
        sa.Column("project_root_data_id", sa.Integer(), nullable=True),
        sa.ForeignKeyConstraint(
            ["project_id"],
            ["project.id"],
            name=op.f("fk_project_root_project_id_project"),
        ),
        sa.ForeignKeyConstraint(
            ["project_root_data_id"],
            ["project_data.id"],
            name=op.f("fk_project_root_project_root_data_id_project_data"),
        ),
        sa.PrimaryKeyConstraint("project_id", name=op.f("pk_project_root")),
    )
    op.create_table(
        "project_shared_data",
        sa.Column("project_id", sa.Integer(), nullable=False),
        sa.Column("project_data_id", sa.Integer(), nullable=False),
        sa.ForeignKeyConstraint(
            ["project_data_id"],
            ["project_data.id"],
            name=op.f("fk_project_shared_data_project_data_id_project_data"),
        ),
        sa.ForeignKeyConstraint(
            ["project_id"],
            ["project.id"],
            name=op.f("fk_project_shared_data_project_id_project"),
        ),
        sa.PrimaryKeyConstraint(
            "project_id",
            "project_data_id",
            name=op.f("pk_project_shared_data"),
        ),
    )
    op.create_table(
        "project_task",
        sa.Column("task_id", sa.String(), nullable=True),
        sa.Column("started_at", sa.DateTime(), nullable=True),
        sa.Column("finished_at", sa.DateTime(), nullable=True),
        sa.Column("state", sa.String(), nullable=True),
        sa.Column("error_message", sa.String(), nullable=True),
        sa.Column(
            "type",
            sa.Enum(
                "process_project_data",
                "process_zone_proposition",
                name="projecttasktype",
            ),
            nullable=True,
        ),
        sa.Column(
            "params", postgresql.JSONB(astext_type=sa.Text()), nullable=True
        ),
        sa.Column("project_id", sa.Integer(), nullable=False),
        sa.Column("id", sa.Integer(), nullable=False),
        sa.ForeignKeyConstraint(
            ["project_id"],
            ["project.id"],
            name=op.f("fk_project_task_project_id_project"),
        ),
        sa.PrimaryKeyConstraint("id", name=op.f("pk_project_task")),
    )
    op.create_table(
        "stream_geo_data_permission",
        sa.Column("object_id", sa.Integer(), nullable=True),
        sa.Column("user_id", sa.Integer(), nullable=False),
        sa.Column("id", sa.Integer(), nullable=False),
        sa.ForeignKeyConstraint(
            ["object_id"],
            ["stream_geo_data.id"],
            name=op.f(
                "fk_stream_geo_data_permission_object_id_stream_geo_data"
            ),
        ),
        sa.ForeignKeyConstraint(
            ["user_id"],
            ["user.id"],
            name=op.f("fk_stream_geo_data_permission_user_id_user"),
        ),
        sa.PrimaryKeyConstraint(
            "id", name=op.f("pk_stream_geo_data_permission")
        ),
    )
    op.create_table(
        "weighted_sum_operand",
        sa.Column("attribute_id", sa.Integer(), nullable=True),
        sa.Column("weight", sa.Float(), nullable=True),
        sa.Column("model_id", sa.Integer(), nullable=False),
        sa.Column("id", sa.Integer(), nullable=False),
        sa.ForeignKeyConstraint(
            ["attribute_id"],
            ["data_attribute.id"],
            name=op.f("fk_weighted_sum_operand_attribute_id_data_attribute"),
            ondelete="SET NULL",
        ),
        sa.ForeignKeyConstraint(
            ["model_id"],
            ["preference_model.id"],
            name=op.f("fk_weighted_sum_operand_model_id_preference_model"),
            ondelete="CASCADE",
        ),
        sa.PrimaryKeyConstraint("id", name=op.f("pk_weighted_sum_operand")),
    )
    op.create_table(
        "wfs_geo_data",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("feature_type", sa.String(), nullable=False),
        sa.Column("version", sa.String(), nullable=True),
        sa.ForeignKeyConstraint(
            ["id"],
            ["stream_geo_data.id"],
            name=op.f("fk_wfs_geo_data_id_stream_geo_data"),
        ),
        sa.PrimaryKeyConstraint("id", name=op.f("pk_wfs_geo_data")),
    )
    op.create_table(
        "wms_geo_data",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column(
            "classes",
            postgresql.ARRAY(sa.Float(), dimensions=2),
            nullable=True,
        ),
        sa.Column("layer", sa.String(), nullable=False),
        sa.Column("resolution", sa.Float(), nullable=True),
        sa.Column("version", sa.String(), nullable=True),
        sa.ForeignKeyConstraint(
            ["id"],
            ["stream_geo_data.id"],
            name=op.f("fk_wms_geo_data_id_stream_geo_data"),
        ),
        sa.PrimaryKeyConstraint("id", name=op.f("pk_wms_geo_data")),
    )
    op.create_table(
        "zone_proposition",
        sa.Column("data_id", sa.Integer(), nullable=True),
        sa.Column("project_id", sa.Integer(), nullable=True),
        sa.Column("id", sa.Integer(), nullable=False),
        sa.ForeignKeyConstraint(
            ["data_id"],
            ["base_geo_data.id"],
            name=op.f("fk_zone_proposition_data_id_base_geo_data"),
        ),
        sa.ForeignKeyConstraint(
            ["project_id"],
            ["project.id"],
            name=op.f("fk_zone_proposition_project_id_project"),
        ),
        sa.PrimaryKeyConstraint("id", name=op.f("pk_zone_proposition")),
    )
    op.create_table(
        "geo_data_permission",
        sa.Column("object_id", sa.Integer(), nullable=True),
        sa.Column("user_id", sa.Integer(), nullable=False),
        sa.Column("id", sa.Integer(), nullable=False),
        sa.ForeignKeyConstraint(
            ["object_id"],
            ["geo_data.id"],
            name=op.f("fk_geo_data_permission_object_id_geo_data"),
        ),
        sa.ForeignKeyConstraint(
            ["user_id"],
            ["user.id"],
            name=op.f("fk_geo_data_permission_user_id_user"),
        ),
        sa.PrimaryKeyConstraint("id", name=op.f("pk_geo_data_permission")),
    )
    op.create_table(
        "mrsort_inference_value",
        sa.Column("mrsort_id", sa.Integer(), nullable=False),
        sa.Column("attribute_id", sa.Integer(), nullable=False),
        sa.Column("alternative_id", sa.Integer(), nullable=False),
        sa.Column(
            "value", postgresql.JSONB(astext_type=sa.Text()), nullable=True
        ),
        sa.ForeignKeyConstraint(
            ["alternative_id"],
            ["mrsort_inference_alternative.id"],
            name=op.f(
                "fk_mrsort_inference_value_alternative_id_mrsort_inference_"
                "alternative"
            ),
            ondelete="CASCADE",
        ),
        sa.ForeignKeyConstraint(
            ["mrsort_id", "attribute_id"],
            ["mrsort_criterion.mrsort_id", "mrsort_criterion.attribute_id"],
            name=op.f("fk_mrsort_inference_value_mrsort_id_mrsort_criterion"),
            ondelete="CASCADE",
        ),
        sa.PrimaryKeyConstraint(
            "mrsort_id",
            "attribute_id",
            "alternative_id",
            name=op.f("pk_mrsort_inference_value"),
        ),
    )
    # ### end Alembic commands ###
    op.execute(
        sa.DDL(
            """
/**
 * Dissolve adjacent geometries if they have the same attributes.
 *
 * Warning: the function deletes the old features before inserting the new ones.
 */
CREATE OR REPLACE FUNCTION dissolve_adjacent (data_id int) RETURNS void AS $$
BEGIN
    -- Use a temporary table to store the attributes
    CREATE TEMP TABLE temp_dissolve AS
    WITH merge AS (
        SELECT f.geom AS geom, json_object_agg(v.attribute_id, v.value) AS props
        FROM feature f INNER JOIN data_value v ON f.id = v.feature_id
        WHERE f.geo_data_id = data_id
        GROUP BY f.id
    )
    -- Dissolve the adjacent geometries based on their attributes
    SELECT ST_SetSRID(ST_UnaryUnion(d.geom), 4326) AS geom, d.props AS props
    FROM (
        SELECT UNNEST(ST_ClusterIntersecting(m.geom)) AS geom, m.props::jsonb AS props
        FROM merge m
        GROUP BY m.props::jsonb
    ) AS d;

    -- Delete old features
    DELETE FROM feature WHERE geo_data_id = data_id;

    -- Insert the new dissolved features
    WITH ins AS (
        INSERT INTO feature (geom, geo_data_id)
        SELECT geom, data_id
        FROM temp_dissolve
        RETURNING id, geom
    )
    INSERT INTO data_value (attribute_id, value, feature_id, type)
    SELECT d.key, d.value, ins.id, a.type
    FROM ins INNER JOIN (
        SELECT t.geom, j.key::int, j.value
        FROM temp_dissolve t, jsonb_each(t.props) j
    ) AS d ON ins.geom = d.geom
    INNER JOIN data_attribute a ON d.key = a.id;

    -- Delete the temporary table
    DROP TABLE temp_dissolve;
END;
$$ LANGUAGE plpgsql;
        """  # noqa: E501
        )
    )
    op.execute(
        sa.DDL(
            """
/******************************************************************************
Copyright (c), Mapbox
All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this
   list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright notice, this
   list of conditions and the following disclaimer in the documentation and/or
   other materials provided with the distribution.
3. Neither the name of the copyright holder nor the names of its contributors may
   be used to endorse or promote products derived from this software without
   specific prior written permission.


THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


https://github.com/mapbox/postgis-vt-util


### TileBBox ###
Given a Web Mercator tile ID as (z, x, y), returns a bounding-box
geometry of the area covered by that tile.
__Parameters:__
- `integer` z - A tile zoom level.
- `integer` x - A tile x-position.
- `integer` y - A tile y-position.
- `integer` srid - SRID of the desired target projection of the bounding
  box. Defaults to 3857 (Web Mercator).
__Returns:__ `geometry(polygon)`
******************************************************************************/
create or replace function TileBBox (z int, x int, y int, srid int = 3857)
    returns geometry
    language plpgsql immutable as
$func$
declare
    max numeric := 20037508.34;
    res numeric := (max*2)/(2^z);
    bbox geometry;
begin
    bbox := ST_MakeEnvelope(
        -max + (x * res),
        max - (y * res),
        -max + (x * res) + res,
        max - (y * res) - res,
        3857
    );
    if srid = 3857 then
        return bbox;
    else
        return ST_Transform(bbox, srid);
    end if;
end;
$func$;
        """  # noqa: E501
        )
    )
    op.execute(
        sa.DDL(
            """
/**
 * Trigger to correct the invalid geometries before an INSERT or an UPDATE.
 */
CREATE OR REPLACE FUNCTION feature_make_valid() RETURNS trigger AS $feature_make_valid$
BEGIN
    IF ST_IsValid(NEW.geom) = 'f' THEN
        NEW.geom := ST_MakeValid(NEW.geom);
    END IF;
    RETURN NEW;
END;
$feature_make_valid$ LANGUAGE plpgsql;

CREATE TRIGGER feature_make_valid BEFORE INSERT OR UPDATE ON feature
    FOR EACH ROW EXECUTE PROCEDURE feature_make_valid();
        """  # noqa: E501
        )
    )


def downgrade():
    op.execute(sa.DDL("DROP TRIGGER feature_make_valid ON feature;"))
    op.execute(sa.DDL("DROP FUNCTION feature_make_valid();"))
    op.execute(sa.DDL("DROP FUNCTION dissolve_adjacent(data_id int);"))
    op.execute(
        sa.DDL("DROP FUNCTION TileBBox(z int, x int, y int, srid int);")
    )

    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table("mrsort_inference_value")
    op.drop_table("geo_data_permission")
    op.drop_table("zone_proposition")
    op.drop_table("wms_geo_data")
    op.drop_table("wfs_geo_data")
    op.drop_table("weighted_sum_operand")
    op.drop_table("stream_geo_data_permission")
    op.drop_table("project_task")
    op.drop_table("project_shared_data")
    op.drop_table("project_root")
    op.drop_table("project_permission")
    op.drop_table("mrsort_inference_alternative")
    op.drop_table("mrsort_criterion")
    op.drop_table("geo_data")
    op.drop_table("default_value")
    with op.batch_alter_table("data_value", schema=None) as batch_op:
        batch_op.drop_index(batch_op.f("ix_data_value_feature_id"))
        batch_op.drop_index(batch_op.f("ix_data_value_attribute_id"))

    op.drop_table("data_value")
    op.drop_table("user_refresh_token")
    op.drop_table("user_permission")
    op.drop_table("stream_geo_data")
    with op.batch_alter_table("project", schema=None) as batch_op:
        batch_op.drop_geospatial_index(
            "idx_project_extent", postgresql_using="gist", column_name="extent"
        )
    op.drop_constraint(
        op.f("fk_project_data_project_id_project"), "project_data"
    )

    op.drop_geospatial_table("project")
    with op.batch_alter_table("feature", schema=None) as batch_op:
        batch_op.drop_index(batch_op.f("ix_feature_geo_data_id"))
        batch_op.drop_geospatial_index(
            "idx_feature_geom", postgresql_using="gist", column_name="geom"
        )

    op.drop_geospatial_table("feature")
    op.drop_table("discrete_category")
    op.drop_table("data_share")
    op.drop_table("data_input_association")
    with op.batch_alter_table("data_attribute", schema=None) as batch_op:
        batch_op.drop_index(batch_op.f("ix_data_attribute_geo_data_id"))

    op.drop_table("data_attribute")
    op.drop_table("user")
    op.drop_constraint(
        op.f("fk_preference_model_data_generator_id_project_data"),
        "preference_model",
    )
    op.drop_table("project_data")
    op.drop_table("preference_model")
    with op.batch_alter_table("base_geo_data", schema=None) as batch_op:
        batch_op.drop_geospatial_index(
            "idx_base_geo_data_extent",
            postgresql_using="gist",
            column_name="extent",
        )

    op.drop_geospatial_table("base_geo_data")
    # ### end Alembic commands ###

    sa.Enum(
        "geo_data",
        "generated_geo_data",
        "proposition_geo_data",
        name="geodatatype",
    ).drop(op.get_bind(), checkfirst=False)
    sa.Enum(
        "categories_rule",
        "continuous_rule",
        "geo_buffer",
        "mrsort",
        "weighted_sum",
        name="preftype",
    ).drop(op.get_bind(), checkfirst=False)
    sa.Enum("all", "min", "max", "sum", "average", name="keepoverlap").drop(
        op.get_bind(), checkfirst=False
    )
    sa.Enum("geo_data", "generator", name="datatype").drop(
        op.get_bind(), checkfirst=False
    )
    sa.Enum("quantitative", "nominal", "ordinal", name="attributetype").drop(
        op.get_bind(), checkfirst=False
    )
    sa.Enum("wfs", "wms", name="streamdatatype").drop(
        op.get_bind(), checkfirst=False
    )
    sa.Enum(
        "create_project", "create_geo_data", name="permissionability"
    ).drop(op.get_bind(), checkfirst=False)
    sa.Enum(
        "process_project_data",
        "process_zone_proposition",
        name="projecttasktype",
    ).drop(op.get_bind(), checkfirst=False)



from collections import defaultdict
from copy import deepcopy

from ..exceptions import PermissionError, RequestError
from ..models import DataAttribute, DataGenerator, DataStream, ProjectData
from .permission_service import has_permission


class ProjectDataService:
    __model__ = ProjectData

    def get_used_input_attributes(self, data_id) -> list[DataAttribute]:
        """Get used input attributes of project data if the user is authorized
        to access it.

        This list contains al inputl attributes used at least once in a
        processing model of the project data.

        Raises:
            PermissionError if the user is not authorized to access the data's
            project.
            RequestError (code 404) if the project data does not exist.

        Returns:
            Used input attributes.
        """
        data = self.get_if_authorized(data_id)
        return data.get_used_input_attributes()

    def get_used_attributes(self, data_id) -> list[DataAttribute]:
        """Get used attributes of project data if the user is authorized
        to access it.

        This list contains all attributes used at least once in a processing
        model of the downstream project data.

        Raises:
            PermissionError if the user is not authorized to access the data's
            project.
            RequestError (code 404) if the project data does not exist.

        Returns:
            Used attributes.
        """
        data = self.get_if_authorized(data_id)
        return data.get_used_attributes()

    def _filter_input_data(self, data):
        """Returns the list of data that need to be processed

        A data needs to be processed if its input data is more recent.
        """

        def recursive_filter(nodes, data, level=0):
            data_process = False
            if data.is_outdated():
                nodes[data.id]
                data_process = True
            for d in data.input_data:
                if isinstance(d, DataGenerator) or isinstance(d, DataStream):
                    if recursive_filter(nodes, d, level=level + 1):
                        nodes[data.id].append(d.id)
                        data_process = True
            return data_process

        nodes = defaultdict(list)
        nodes[data.id]
        recursive_filter(nodes, data)
        return dict(nodes)

    def _topological_sort(self, n):
        """Returns a topological sort of the nodes using Kahn's algorithm"""
        nodes = deepcopy(n)
        sorted_data = []
        set_queue = []
        for node, edges in list(nodes.items()):
            if len(edges) == 0:
                del nodes[node]
                set_queue.append(node)
        while len(set_queue) > 0:
            current = set_queue.pop()
            sorted_data.append(current)
            for node, edges in list(nodes.items()):
                if current in edges:
                    edges.remove(current)
                    if len(edges) == 0:
                        del nodes[node]
                        set_queue.append(node)
        return sorted_data

    def get_grouped_dependant_data_ids(self, data):
        """Return the ids of project data that need to be processed, sorted and
        grouped by depth.

        The grouped project data are independant and can be processed in
        parallel.
        """
        nodes_depth = {}
        nodes = self._filter_input_data(data)
        sorted_data = self._topological_sort(nodes)
        for node in sorted_data:
            depths = [0]
            for src in nodes[node]:
                if src in nodes_depth.keys():
                    depths.append(nodes_depth[src] + 1)
            nodes_depth[node] = max(depths)
        result = []
        for current_depth in set(sorted(nodes_depth.values())):
            result.append([])
            for node, depth in list(nodes_depth.items()):
                if depth == current_depth:
                    del nodes_depth[node]
                    result[-1].append(node)
        return result

    def get_if_authorized(self, project_data_id) -> ProjectData:
        """Load the the project data if the user is authorized to access it.

        Raises:
            PermissionError if the user is not authorized to access the data's
            project.
            RequestError (code 404) if the project data does not exist.

        Returns:
            The project data object.
        """
        project_data = self.__model__.get_by_id(project_data_id)
        if project_data is None:
            raise RequestError(
                "The project data requested does not exist", 404
            )
        if not has_permission(project_data.project):
            raise PermissionError
        return project_data



from . import create_app
from .cli import init_app as init_cli
from .extensions import init_app as init_extensions

# Following is necessary for worker to be launched with
# celery -A deseasion.backend.app.celery worker
from .extensions.celery import celery  # noqa: F401

app = create_app("instance/config.py")
init_extensions(app)
init_cli(app)


if __name__ == "__main__":
    app.run(host="0.0.0.0", port=5000, debug=True)



from celery import Celery
from flask import Flask

from ..tasks import ContextTask

celery = Celery()


def init_app(app: Flask):
    celery.main = app.import_name
    celery.conf.result_backend = app.config["CELERY_RESULT_BACKEND"]
    celery.conf.broker_url = app.config["CELERY_BROKER_URL"]
    celery.conf.update(app.config)

    ContextTask.app_context = app.app_context()
    celery.Task = ContextTask



"""Add extent_filter to data

Revision ID: 337e01e5199e
Revises: bb99ff70c7bf
Create Date: 2025-06-26 12:17:46.133271

"""

import sqlalchemy as sa
from alembic import op
from geoalchemy2 import Geometry

# revision identifiers, used by Alembic.
revision = "337e01e5199e"
down_revision = "bb99ff70c7bf"
branch_labels = None
depends_on = None


def upgrade():
    with op.batch_alter_table("base_data", schema=None) as batch_op:
        batch_op.add_geospatial_column(
            sa.Column(
                "extent_filter",
                Geometry(
                    spatial_index=False,
                    from_text="ST_GeomFromEWKT",
                    name="geometry",
                ),
                nullable=True,
            )
        )
        batch_op.create_geospatial_index(
            "idx_base_data_extent_filter",
            ["extent_filter"],
            unique=False,
            postgresql_using="gist",
            postgresql_ops={},
        )


def downgrade():
    with op.batch_alter_table("base_data", schema=None) as batch_op:
        batch_op.drop_geospatial_index(
            "idx_base_data_extent_filter",
            postgresql_using="gist",
            column_name="extent_filter",
        )
        batch_op.drop_geospatial_column("extent_filter")



"""Split disjointed geometries.

Revision ID: a212e7950484
Revises: 184b71fa9bdc
Create Date: 2025-03-20 17:13:00.000000

"""

import sqlalchemy as sa
from alembic import op
from geoalchemy2 import Geometry, WKBElement, shape
from sqlalchemy.dialects import postgresql

from deseasion.backend.services.geo_data_loading_service import (
    split_multi_geometry,
)
from deseasion.backend.services.utils import progress_bar

# revision identifiers, used by Alembic.
revision = "a212e7950484"
down_revision = "184b71fa9bdc"
branch_labels = None
depends_on = None


feature = sa.Table(
    "feature",
    sa.MetaData(),
    sa.Column(
        "type",
        sa.Enum("feature", "geo_feature", name="featuretype"),
        nullable=True,
    ),
    sa.Column("data_id", sa.Integer(), nullable=True),
    sa.Column(
        "explainability",
        postgresql.JSONB(astext_type=sa.Text()),
        nullable=True,
    ),
    sa.Column("id", sa.Integer(), nullable=False),
)

geo_feature = sa.Table(
    "geo_feature",
    sa.MetaData(),
    sa.Column("id", sa.Integer(), nullable=False),
    sa.Column(
        "geom",
        Geometry(
            spatial_index=False, from_text="ST_GeomFromEWKT", name="geometry"
        ),
        nullable=True,
    ),
)

data_value = sa.Table(
    "data_value",
    sa.MetaData(),
    sa.Column(
        "type",
        sa.Enum("quantitative", "nominal", "ordinal", name="attributetype"),
        nullable=True,
    ),
    sa.Column("attribute_id", sa.Integer(), nullable=True),
    sa.Column("feature_id", sa.Integer(), nullable=True),
    sa.Column("value", postgresql.JSONB(astext_type=sa.Text()), nullable=True),
    sa.Column("id", sa.Integer(), nullable=False),
)


feature_input_assoc = sa.Table(
    "feature_input_association",
    sa.MetaData(),
    sa.Column(
        "feature_id",
        sa.Integer,
    ),
    sa.Column(
        "input_feature_id",
        sa.Integer,
    ),
)


def upgrade():
    conn = op.get_bind()
    # Get all geo features with non-polygon geometry
    nb = conn.execute(
        sa.text(
            "SELECT count(id) FROM geo_feature "
            "WHERE ST_GeometryType(geom) != 'ST_Polygon'"
        )
    ).fetchone()[0]
    res = conn.execute(
        sa.text(
            "SELECT f.id, g.geom, f.data_id, f.explainability "
            "FROM geo_feature AS g JOIN feature AS f ON f.id = g.id "
            "WHERE ST_GeometryType(g.geom) != 'ST_Polygon'"
        )
    )
    i = 0
    progress_bar(i, nb)
    for id, geom, data_id, explain in res:
        # Split disjoint geometry
        geoms = split_multi_geometry(shape.to_shape(WKBElement(geom)))
        # Update old geometry with first single part
        geom1 = next(geoms)
        conn.execute(
            geo_feature.update()
            .values(geom=shape.from_shape(geom1, srid=4326))
            .where(geo_feature.c.id == id)
        )
        # Insert a new geo feature for each other simple geometry
        for simple_geom in geoms:
            new_id = conn.execute(
                feature.insert()
                .values(
                    data_id=data_id, explainability=explain, type="geo_feature"
                )
                .returning(feature.c.id)
            ).first()[0]
            conn.execute(
                geo_feature.insert().values(
                    id=new_id, geom=shape.from_shape(simple_geom, srid=4326)
                )
            )
            # Duplicate each data value for new feature
            sel = sa.select(
                data_value.c["type", "attribute_id", "value"],
                sa.bindparam("feature_id", new_id),
            ).where(data_value.c.feature_id == id)
            conn.execute(
                data_value.insert().from_select(
                    ["type", "attribute_id", "value", "feature_id"], sel
                )
            )
            # Add input features of old feature
            sel = sa.select(
                feature_input_assoc.c["input_feature_id"],
                sa.bindparam("feature_id", new_id),
            ).where(feature_input_assoc.c.feature_id == id)
            conn.execute(
                feature_input_assoc.insert().from_select(
                    ["input_feature_id", "feature_id"], sel
                )
            )
            # Add new feature as input of features having old feature as input
            # Old feature was union of new features
            sel = sa.select(
                feature_input_assoc.c["feature_id"],
                sa.bindparam("input_feature_id", new_id),
            ).where(feature_input_assoc.c.input_feature_id == id)
            conn.execute(
                feature_input_assoc.insert().from_select(
                    ["feature_id", "input_feature_id"], sel
                )
            )
        i += 1
        progress_bar(i, nb)


def downgrade():
    pass



DEBUG = False

PROPAGATE_EXCEPTIONS = True  # allow the app to handle exceptions

# SQLAlchemy configuration
SQLALCHEMY_DATABASE_URI = ""  # overwrite this option in `instance/config.py`
SQLALCHEMY_TRACK_MODIFICATIONS = False

# Swagger configuration (for flasgger)
SWAGGER = {
    "title": "DESEASION API docs",
    "version": "2.1.1",
    "openapi": "3.0.2",
}
API_ROOT = "/api"



"""replace project data shares with data shares

Revision ID: 163a9f52cac3
Revises: c6e3da96b784
Create Date: 2025-06-13 14:23:25.338030

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "163a9f52cac3"
down_revision = "c6e3da96b784"
branch_labels = None
depends_on = None


def upgrade():
    # Remove foreign key constraint on data_share.data_id
    with op.batch_alter_table("data_share", schema=None) as batch_op:
        batch_op.drop_constraint(
            "fk_data_share_data_id_project_data", type_="foreignkey"
        )

    op.execute(
        sa.text(
            """
UPDATE data_share
SET data_id = project_data.data_id
FROM project_data
WHERE data_share.data_id = project_data.id;
"""
        )
    )
    op.execute(sa.text("DELETE FROM data_share WHERE data_id IS NULL"))

    # Add updated foreign key constraint on data_share.data_id
    with op.batch_alter_table("data_share", schema=None) as batch_op:
        batch_op.create_foreign_key(
            batch_op.f("fk_data_share_data_id_base_data"),
            "base_data",
            ["data_id"],
            ["id"],
        )


def downgrade():
    # Remove foreign key constraint on data_share.data_id
    with op.batch_alter_table("data_share", schema=None) as batch_op:
        batch_op.drop_constraint(
            batch_op.f("fk_data_share_data_id_base_data"), type_="foreignkey"
        )

    op.execute(
        sa.text(
            """
DELETE FROM data_share WHERE id IN
(
    SELECT s.id FROM data_share s
    JOIN base_data d ON s.data_id = d.id
    WHERE d.type != 'generated_geo_data'
)
"""
        )
    )
    op.execute(
        sa.text(
            """
UPDATE data_share
SET data_id = project_data.id
FROM project_data
WHERE data_share.data_id = project_data.data_id;
"""
        )
    )
    op.execute(sa.text("DELETE FROM data_share WHERE data_id IS NULL"))

    # Add updated foreign key constraint on data_share.data_id
    with op.batch_alter_table("data_share", schema=None) as batch_op:
        batch_op.create_foreign_key(
            batch_op.f("fk_data_share_data_id_project_data"),
            "project_data",
            ["data_id"],
            ["id"],
        )



"""Remove project shared data

Revision ID: 9a7260e6a3bf
Revises: 2ce65db29472
Create Date: 2024-11-21 14:07:50.542424

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "9a7260e6a3bf"
down_revision = "4ca4e5e44ff2"
branch_labels = None
depends_on = None


project_data = sa.Table(
    "project_data",
    sa.MetaData(),
    sa.Column(
        "data_type",
        sa.Enum("geo_data", "generator", "global_data", name="datatype"),
        nullable=True,
    ),
    sa.Column("id", sa.Integer(), nullable=False),
    sa.Column("data_id", sa.Integer(), nullable=True),
    sa.Column("last_update", sa.DateTime(), nullable=True),
    sa.Column("project_id", sa.Integer(), nullable=False),
    sa.Column("description", sa.String(), nullable=True),
    sa.Column("name", sa.String(), nullable=True),
    sa.Column("active_model_id", sa.Integer(), nullable=True),
)


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table("project_shared_data")

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    project_shared_data = op.create_table(
        "project_shared_data",
        sa.Column(
            "project_id", sa.INTEGER(), autoincrement=False, nullable=False
        ),
        sa.Column(
            "project_data_id",
            sa.INTEGER(),
            autoincrement=False,
            nullable=False,
        ),
        sa.ForeignKeyConstraint(
            ["project_data_id"],
            ["project_data.id"],
            name="fk_project_shared_data_project_data_id_project_data",
        ),
        sa.ForeignKeyConstraint(
            ["project_id"],
            ["project.id"],
            name="fk_project_shared_data_project_id_project",
        ),
        sa.PrimaryKeyConstraint(
            "project_id", "project_data_id", name="pk_project_shared_data"
        ),
    )

    shared_data = sa.select(project_data.c["project_id", "id"]).where(
        project_data.c.data_type != "generator"
    )
    op.execute(
        project_shared_data.insert().from_select(
            names=["project_id", "project_data_id"], select=shared_data
        )
    )
    # ### end Alembic commands ###



import random

import pulp as pp

from ..exceptions import InferenceError, RequestError
from ..models.geo_data import DataAttribute
from ..models.geo_data_utils import ModelProcessingService
from ..models.mrsort_inference import MRSortInferenceAlternative
from ..models.processing_models import ProcessingModel
from ..tasks import process_and_set_model, update_stream_data


class ProcessingModelService:
    def get_used_input_attributes(self, model_id) -> list[DataAttribute]:
        """Get list of used input attributes in a model definition.

        :param model_id:
        :raises RequestError: 404, if model `model_id` does not exist.
        :return:
        """
        model = ProcessingModel.get_by_id(model_id)
        if model is None:
            raise RequestError("The model does not exist", 404)
        return model.get_used_input_attributes()

    def task_process(self, model_id):
        """
        Start a celery task to process the model.

        Args:
            model_id: The id of the processing model.

        Returns:
            A celery task.
        """
        return process_and_set_model.si(model_id=model_id)

    def update_stream_process(
        self,
        data_id,
    ):
        return update_stream_data.si(
            data_id=data_id,
        )

    def _get_mrsort_alternatives(self, model, data, count=100):
        """
        Choose alternatives for the inference model.

        Args:
            model: A MR-Sort model.
            data: A list of the data for each attribute in the form
                  [{'attr1': val, 'attr2': val}, {'attr1': val, ...}, ...]
            count (int): The number of alternatives to propose. Default: 100.
        """
        # Make sure the criteria are up to date
        model.init_criteria()
        criteria = {c.attribute: c for c in model.criteria}
        choices = []
        for d in data:
            # Filter duplicates and ignore when the value of an attribute is
            # None
            if d not in choices and all((v is not None for v in d.values())):
                choices.append(d)
        try:
            final_choices = random.sample(choices, count)
        except ValueError:
            final_choices = choices
        alternatives = []
        for c in final_choices:
            alternative = MRSortInferenceAlternative(mrsort=model)
            for attr, val in c.items():
                # Only consider model criteria (i.e quantitative data)
                if attr in criteria:
                    alternative.values[criteria[attr]] = val
            alternatives.append(alternative)
        return alternatives

    def load_mrsort_inference_data(self, model):
        """
        Decompose the geometries to generate the alternatives for the MR-Sort
        inference.
        """
        processing_service = ModelProcessingService()
        data = processing_service.get_decomposed_data(model)
        alternatives = self._get_mrsort_alternatives(model, data)
        model.inference_alternatives = alternatives
        model.update()
        return alternatives

    def _solve_lp(self, X, F, Fdir, Fmin, Fmax, K, A, PTx, gamma):
        """
        Solve the linear problem for the inference of the MR-Sort model.

        Args:
            X: Number of alternatives.
            F: Number of criteria.
            Fdir:
                Criteria perference directions
                (1 for maximizing, -1 for minimizing).
            Fmin: Minimum values for the criteria.
            Fmax: Maximum values for the criteria.
            K: Number of categories.
            A: Assignements for the alternatives (0 is best, K-1 is worst).
            PTx: Performance table for the alternatives.
            gamma: Separation value for the inequalities and parameters.
        """
        # Convert the list to a dict
        PTx = {(x, i): PTx[x][i] for x in range(X) for i in range(F)}

        # Initialize the problem
        prob = pp.LpProblem("MR-Sort inference", pp.LpMinimize)

        # Variables
        _lambda = pp.LpVariable("majority threshold", 0.5, 1)
        w = pp.LpVariable.dicts("weight", range(F), 0, 1)
        PTk = pp.LpVariable.dicts(
            "performance profiles",
            ((k, i) for k in range(K + 1) for i in range(F)),
        )
        lCupp = pp.LpVariable.dicts(
            "upper profile",
            ((x, i) for x in range(X) for i in range(F)),
            cat=pp.LpBinary,
        )
        lClow = pp.LpVariable.dicts(
            "lower profile",
            ((x, i) for x in range(X) for i in range(F)),
            cat=pp.LpBinary,
        )
        wlCupp = pp.LpVariable.dicts(
            "weighted upper", ((x, i) for x in range(X) for i in range(F))
        )
        wlClow = pp.LpVariable.dicts(
            "weighted lower", ((x, i) for x in range(X) for i in range(F))
        )

        # Constraints
        prob += pp.lpSum(w) == 1, "normalizedWeights"
        for i in range(F):
            prob += (
                PTk[K, i] == Fmin[i] - gamma
                if Fdir[i] > 0
                else Fmax[i] + gamma
            )
            prob += (
                PTk[0, i] == Fmax[i] + gamma
                if Fdir[i] > 0
                else Fmin[i] - gamma
            )
            for x in range(X):
                f_diff = Fmax[i] - Fmin[i] + 1
                prob += (
                    Fdir[i] * (PTx[x, i] - PTk[A[x], i]) + gamma
                    <= lCupp[x, i] * f_diff
                )
                prob += (lCupp[x, i] - 1) * f_diff <= Fdir[i] * (
                    PTx[x, i] - PTk[A[x], i]
                )
                prob += (
                    Fdir[i] * (PTx[x, i] - PTk[A[x] + 1, i]) + gamma
                    <= lClow[x, i] * f_diff
                )
                prob += (lClow[x, i] - 1) * f_diff <= Fdir[i] * (
                    PTx[x, i] - PTk[A[x] + 1, i]
                )
                prob += wlCupp[x, i] <= w[i]
                prob += 0 <= wlCupp[x, i]
                prob += wlCupp[x, i] <= lCupp[x, i]
                prob += lCupp[x, i] + w[i] - 1 <= wlCupp[x, i]
                prob += wlClow[x, i] <= w[i]
                prob += 0 <= wlClow[x, i]
                prob += wlClow[x, i] <= lClow[x, i]
                prob += lClow[x, i] + w[i] - 1 <= wlClow[x, i]
            for k in range(K):
                prob += Fdir[i] * PTk[(k + 1), i] <= Fdir[i] * PTk[k, i]
        for x in range(X):
            prob += _lambda <= pp.lpSum([wlClow[x, i] for i in range(F)])
            prob += (
                pp.lpSum([wlCupp[x, i] for i in range(F)]) + gamma <= _lambda
            )

        # Solve the problem
        # use prob.solve(pp.GLPK(msg=False)) to solve with GLPK
        prob.solve()

        if pp.LpStatus[prob.status] != "Optimal":
            raise InferenceError()

        result = {
            "lambda": _lambda.varValue,
            "weights": [w[i].varValue for i in range(F)],
            "profiles": [
                [PTk[k, i].varValue for i in range(F)] for k in range(K + 1)
            ],
        }
        return result

    def infer_model(self, mrsort):
        """
        Infer the MR-Sort model and update its parameters from the
        alternatives.

        Args:
            mrsort: The MR-Sort object.
        """
        alternatives = list(
            filter(
                lambda a: a.category is not None, mrsort.inference_alternatives
            )
        )
        categories = list(mrsort.categories)
        model = self._solve_lp(
            X=len(alternatives),
            F=len(mrsort.criteria),
            Fdir=[1 if c.maximize else -1 for c in mrsort.criteria],
            Fmin=[
                min((v.value for v in c.inference_values))
                for c in mrsort.criteria
            ],
            Fmax=[
                max((v.value for v in c.inference_values))
                for c in mrsort.criteria
            ],
            K=len(mrsort.categories),
            A=[categories.index(a.category) for a in alternatives],
            PTx=[[a.values[c] for c in mrsort.criteria] for a in alternatives],
            gamma=0.1,
        )

        majority_threshold = model.get("lambda")
        weights = model.get("weights", [])
        profiles = model.get("profiles", [[]])

        mrsort.majority_threshold = majority_threshold
        for index, criterion in enumerate(mrsort.criteria):
            criterion.maximize = mrsort.criteria[index].maximize
            criterion.weight = weights[index]
            prof = [p[index] for p in profiles]
            criterion.profiles = prof[
                len(prof) - 2 : 0 : -1
            ]  # reverse order (worst first)
        mrsort.update()
